// Copyright (C) 2010 By Paul Dunn
//
// This file is part of the SpecBAS BASIC Interpreter, which is in turn
// part of the SpecOS project.
//
// SpecBAS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// SpecBAS is distributed in the hope that it will be entertaining,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SpecBAS.  If not, see <http://www.gnu.org/licenses/>.

unit MainForm;

{$INCLUDE SpecBAS.inc}

interface

uses
  SHFolder, Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, FastIMG, SP_SysVars, SP_Graphics, SP_Graphics32, SP_BankManager, SP_Util, SP_Main, SP_FileIO,
  ExtCtrls, SP_Input, MMSystem, FastDIB, SP_Errors, SP_Sound, Bass, GraphicEx, SP_Tokenise;

type
  TMain = class(TForm)
    FastIMG1: TFastIMG;
    Timer1: TTimer;
    procedure FormShow(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
    procedure FormDestroy(Sender: TObject);
    procedure FastIMG1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure FastIMG1MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FastIMG1MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure Timer1Timer(Sender: TObject);
    procedure FormMouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
    procedure FormMouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
  private
    { Private declarations }
    procedure OnAppMessage(var Msg: TMsg; var Handled: Boolean);
    Procedure OnIdle(Sender: TObject; Var Done: Boolean);
    procedure CMDialogKey( Var msg: TCMDialogKey ); message CM_DIALOGKEY;
  public
    { Public declarations }
    Function GetCharFromVirtualKey(Key: Word): astring;
  end;

  TSpecBAS_Thread = Class(TThread)
    Procedure Execute; Override;
  End;

  Function  SetScreen(Width, Height: Integer; FullScreen: Boolean): Integer;
  Function  SetScreenResolution(Width, Height: Integer; FullScreen: Boolean): Boolean;
  Function  TestScreenResolution(Width, Height: Integer; FullScreen: Boolean): Boolean;
  Procedure YieldProc;
  Procedure GetKeyState;
  Procedure Refresh_Display;
  Function  GetTicks: LongWord;
  Procedure Quit;
  function  Sto_GetFmtFileVersion(const FileName: String = ''; const Fmt: String = '%d.%d'): String;
  Procedure LoadImage(Filename: aString; Var Error: TSP_ErrorCode);
  Procedure FreeImageResource;

var
  Main: TMain;
  RealScreenWidth, RealScreenHeight, ScrWidth, ScrHeight: Integer;
  BASThread: TSpecBAS_Thread;
  Quitting: Boolean = False;
  OldTicks: LongWord = 0;
  InitTime: LongWord;
  ImgResource: TFastDIB;
  ThreadAlive: Boolean = False;

implementation

{$R *.dfm}

Procedure TMain.OnIdle(Sender: TObject; Var Done: Boolean);
Var
  Update: Boolean;
  Ticks: LongWord;
Begin

  Update := True;
  Ticks := TimeGetTime;
  Done := PROGSTATE <> SP_PR_RUN;
  While Ticks - OldTicks >= LongWord(FRAME_MS) Do Begin
    Inc(OldTicks, FRAME_MS);
    Inc(FRAMES);
    SP_FrameUpdate;
    Ticks := TimeGetTime;
    If CauseUpdate And Update Then Begin
      CB_Refresh_Display;
      CauseUpdate := False;
      Update := False;
      Done := False;
    End;
  End;

  If Not Done Then CB_YIELD;

End;

procedure TMain.Timer1Timer(Sender: TObject);
begin

  // Nowt.

end;

Function GetTicks: LongWord;
Begin

  Result := TimeGetTime;

End;

Function SetScreen(Width, Height: Integer; FullScreen: Boolean): Integer;
Begin

  Result := 0;
  DISPLAYPOINTER := nil;
  DPtrBackup := Nil;

  If FullScreen Then
    Main.BorderStyle := BsNone
  Else
    Main.BorderStyle := BsSingle;

  SetScreenResolution(Width, Height, FullScreen);

  Main.ClientWidth := Width;
  Main.ClientHeight := Height;
  Main.Left := (Screen.Width - Width) Div 2;
  Main.Top := (Screen.Height - Height) Div 2;
  If Not Quitting Then Begin
    Main.FastIMG1.Bmp.SetSize(Width, -Height, 32);
    DISPLAYSTRIDE := Main.FastIMG1.Bmp.BWidth;
    DISPLAYPOINTER := Main.FastIMG1.Bmp.Bits;
    DPtrBackup := DISPLAYPOINTER;
  End;

End;

function TestScreenResolution(Width, Height: Integer; FullScreen: Boolean): Boolean;
var
  DeviceMode: TDeviceMode;
begin

  If FullScreen Then Begin

    with DeviceMode do begin
      dmSize := SizeOf(TDeviceMode);
      dmPelsWidth := Width;
      dmPelsHeight := Height;
      dmFields := DM_PELSWIDTH or DM_PELSHEIGHT;
    end;
    Result := ChangeDisplaySettings(DeviceMode, CDS_TEST) = DISP_CHANGE_SUCCESSFUL;

  End Else
    Result := True;

end;

function SetScreenResolution(Width, Height: Integer; FullScreen: Boolean): Boolean;
var
  DeviceMode: TDeviceMode;
begin

  If FullScreen Then Begin
    with DeviceMode do begin
      dmSize := SizeOf(TDeviceMode);
      dmPelsWidth := Width;
      dmPelsHeight := Height;
      dmFields := DM_PELSWIDTH or DM_PELSHEIGHT;
    end;
    Result := ChangeDisplaySettingsEx(Nil, DeviceMode, 0, CDS_FULLSCREEN, nil) = DISP_CHANGE_SUCCESSFUL;;
    SPFULLSCREEN := True;
  End Else Begin
    If SPFULLSCREEN Then Begin
      with DeviceMode do begin
        dmSize := SizeOf(TDeviceMode);
        dmPelsWidth := RealScreenWidth;
        dmPelsHeight := RealScreenHeight;
        dmFields := DM_PELSWIDTH or DM_PELSHEIGHT;
      end;
      ChangeDisplaySettingsEx(Nil, DeviceMode, 0, CDS_FULLSCREEN, nil);
      SPFULLSCREEN := False;
    End;
    Result := True;
  End;

end;

procedure TMain.OnAppMessage(var Msg: TMsg; var Handled: Boolean);
begin

  case Msg.message of
     WM_KEYDOWN:
      begin
        if (Msg.lParam shr 30) = 1 then begin
          Handled := True;
        end else
          Handled := False;
      end;
  else
     // Not handled
     Handled := False;
  end;

End;


Procedure TSpecBAS_Thread.Execute;
Begin

  ThreadAlive := True;
  Priority := tpTimeCritical;
  SP_MainLoop;
  ThreadAlive := False;
  Terminate;

End;

procedure TMain.FastIMG1MouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin

  MOUSEBTN := Integer(ssLeft in Shift) + (2 * Integer(ssRight in Shift)) + (4 * Integer(ssMiddle in Shift));
  MOUSEX := X;
  MOUSEY := Y;
  M_DOWNFLAG := True;

end;

procedure TMain.FastIMG1MouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin

  M_DELTAX := X - MOUSEX;
  M_DELTAY := Y - MOUSEY;
  M_MOVEFLAG := True;

  MOUSEX := X;
  MOUSEY := Y;

  MOUSEBTN := Integer(ssLeft in Shift) + (2 * Integer(ssRight in Shift)) + (4 * Integer(ssMiddle in Shift));

  // Ensure the mouse pointer is drawn at the new position - won't happen if the screen is locked.

  If MOUSEVISIBLE Then Begin
    SP_RestoreMouseRegion;
    SP_DrawMouseImage;
    SP_NeedDisplayUpdate := True;
  End;

end;

procedure TMain.FastIMG1MouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin

  MOUSEBTN := Integer(ssLeft in Shift) + (2 * Integer(ssRight in Shift)) + (4 * Integer(ssMiddle in Shift));
  MOUSEX := X;
  MOUSEY := Y;
  M_UPFLAG := True;

end;

procedure TMain.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
Var
  Error: TSP_ErrorCode;
begin

  If Not QUITMSG Then Begin

    Error.Code := SP_ERR_OK;
    DoAutoSave(Error);
    Quitting := True;

  End;

  QUITMSG := True;
  Repeat
    Sleep(0);
  Until Not Drawing;
  SP_DisableDisplay;

end;

procedure TMain.FormDestroy(Sender: TObject);
begin

  Quitting := True;
  TimeEndPeriod(TIMERES);
  BASS_Free;

end;

Function TMain.GetCharFromVirtualKey(Key: Word): astring;
var
  keyboardState: TKeyboardState;
  asciiResult: Integer;
  Idx: Integer;
begin

  GetKeyboardState(keyboardState);

  // Filter out the CTRL key - it'll be picked up later.

  KeyBoardState[17] := 0;

  SetLength(Result, 2);
  asciiResult := ToAscii(key, MapVirtualKey(key, 0), KeyboardState, @Result[1], 0);
  case asciiResult of
    0: If Key < 32 Then Result := aChar(Key) Else Result := '';
    1: SetLength(Result, 1);
    2: Result := '';
    else
      Result := '';
  end;

end;


procedure TMain.CMDialogKey(var msg: TCMDialogKey);
begin
  if msg.Charcode <> VK_TAB then
    inherited;
end;

procedure TMain.FormKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
Var
  aStr: aString;
  Err: TSP_ErrorCode;
begin

  K_DOWNFLAG := True;
  if key <> 17 then
    K_DOWNFLAG := false;
  aStr := aString(GetCharFromVirtualKey(Key and $FF));
  If aStr = '' Then aStr := #0;
  LASTKEYCHAR := aStr;
  SP_KeyDown(Key);

end;

procedure TMain.FormKeyUp(Sender: TObject; var Key: Word; Shift: TShiftState);
begin

  K_UPFLAG := True;
  LASTKEYCHAR := #0;
  SP_KeyUp(Key);

end;

procedure TMain.FormMouseWheelDown(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin

  M_WHEELDNFLAG := True;
  Inc(MOUSEWHEEL);

end;

procedure TMain.FormMouseWheelUp(Sender: TObject; Shift: TShiftState; MousePos: TPoint; var Handled: Boolean);
begin

  M_WHEELUPFLAG := True;
  Dec(MouseWheel);

end;

procedure TMain.FormShow(Sender: TObject);
Var
  Path: Array [0..MAX_PATH] of Char;
begin

  FastIMG1.Bmp.RMask := $00FF0000;
  FastIMG1.Bmp.GMask := $0000FF00;
  FastIMG1.Bmp.BMask := $000000FF;
  FastIMG1.Bmp.SetSize(FastIMG1.Width, FastIMG1.Height, FastIMG1.Bmp.Bpp);

  OldTicks := TimeGetTime;
  FastIMG1.Cursor := CrNone;

  TIMERES := 1;
  While TimeBeginPeriod(TIMERES) <> TIMERR_NOERROR Do Inc(TIMERES);
  InitTime := TimeGetTime;

  BUILDSTR := aString(Sto_GetFmtFileVersion('', '%d.%d.%d.%d'));
  SHGetFolderPath(0,$0028,0,SHGFP_TYPE_CURRENT,@path[0]);
  HOMEFOLDER := Path + aString('\specbas');

  If Not DirectoryExists(String(HOMEFOLDER)) Then
    CreateDir(String(HOMEFOLDER));
  If Not DirectoryExists(String(HOMEFOLDER) + '\temp') Then
    CreateDir(String(HOMEFOLDER) + '\temp');
  TEMPDIR := HOMEFOLDER + '\temp\';

  SetCurrentDir(String(HOMEFOLDER));
  HOMEFOLDER := Lower(HOMEFOLDER);
  If HOMEFOLDER[Length(HOMEFOLDER)] <> '\' Then
    HOMEFOLDER := HOMEFOLDER + '\';

  AUTOSAVE := True;

  ScrWidth := 800;
  ScrHeight := 480;
  RealScreenWidth := Screen.Width;
  RealScreenHeight := Screen.Height;

  Application.OnActivate := OnActivate;
  Application.OnDeactivate := OnDeactivate;

  // Initialise callbacks

  CB_GetKeyLockState := GetKeyState;
  CB_Refresh_Display := Refresh_Display;
  CB_Quit := MainForm.Quit;
  CB_SetScreenRes := SetScreen;
  CB_Test_Resolution := TestScreenResolution;
  CB_GetTicks := GetTicks;
  CB_Yield := YieldProc;
  CB_Load_Image := LoadImage;
  CB_Free_Image := FreeImageResource;

  SP_InitialGFXSetup(ScrWidth, ScrHeight);

  // Launch the interpreter

  SP_CLS(CPAPER);
  EDITLINE := '';
  CURSORPOS := 0;
  CURSORCHAR := 32;
  SYSTEMSTATE := SS_IDLE;

  SP_Init_Sound;

  BASThread := TSpecBAS_Thread.Create(False);
  SetThreadAffinityMask(GetCurrentThread(), 0);
  SetThreadAffinityMask(BASThread.ThreadID, 1);

  Application.OnIdle := OnIdle;
  Application.OnMessage := OnAppMessage;

end;

Procedure Quit;
Begin

  Quitting := True;
  SendMessage(Main.Handle, WM_CLOSE, 0, 0);

End;

Procedure GetKeyState;
Begin

  CAPSLOCK := Windows.GetKeyState(VK_CAPITAL) and 1;
  NUMLOCK := WIndows.GetKeyState(VK_NUMLOCK) and 1;

End;

Procedure YieldProc; inline;
Begin

  Sleep(1);

End;

Procedure Refresh_Display;
Var
  X1, Y1, X2, Y2: Integer;
Begin

  DRAWING := True;
  If Not (Quitting or SCREENCHANGE) Then Begin
    If DISPLAYPOINTER <> nil Then Begin
      SP_RestoreMouseRegion;
      If NUMSPRITES > 0 Then Begin
        SCMINX := 0; SCMINY := 0;
        SCMAXX := DISPLAYWIDTH; SCMAXY := DISPLAYHEIGHT;
      End;
      If SCMAXX >= SCMINX Then Begin
        While SetDR Do Sleep(0); SetDR := True;
        X1 := SCMINX; Y1 := SCMINY; X2 := SCMAXX +1; Y2 := SCMAXY +1;
        SCMAXX := 0; SCMAXY := 0; SCMINX := DISPLAYWIDTH; SCMINY := DISPLAYHEIGHT;
        SetDR := False;
        SP_Composite32(DISPLAYPOINTER, X1, Y1, X2, Y2);
      End;
      If MOUSEVISIBLE Then SP_DrawMouseImage;
      Main.FastIMG1.Repaint;
    End;
  End;
  DRAWING := False;

End;

function Sto_GetFmtFileVersion(const FileName: String = ''; const Fmt: String = '%d.%d'): String;
var
  sFileName: String;
  iBufferSize: DWORD;
  iDummy: DWORD;
  pBuffer: Pointer;
  pFileInfo: Pointer;
  iVer: array[1..4] of Word;
begin
  // set default value
  Result := '';
  // get filename of exe/dll if no filename is specified
  sFileName := Trim(FileName);
  if (sFileName = '') then
    sFileName := GetModuleName(HInstance);
  // get size of version info (0 if no version info exists)
  iBufferSize := GetFileVersionInfoSize(PChar(sFileName), iDummy);
  if (iBufferSize > 0) then
  begin
    GetMem(pBuffer, iBufferSize);
    try
    // get fixed file info (language independent)
    GetFileVersionInfo(PChar(sFileName), 0, iBufferSize, pBuffer);
    VerQueryValue(pBuffer, '\', pFileInfo, iDummy);
    // read version blocks
    iVer[1] := HiWord(PVSFixedFileInfo(pFileInfo)^.dwFileVersionMS);
    iVer[2] := LoWord(PVSFixedFileInfo(pFileInfo)^.dwFileVersionMS);
    iVer[3] := HiWord(PVSFixedFileInfo(pFileInfo)^.dwFileVersionLS);
    iVer[4] := LoWord(PVSFixedFileInfo(pFileInfo)^.dwFileVersionLS);
    finally
      FreeMem(pBuffer);
    end;
    // format result string
    Result := IntToStr(iVer[3])+ '.' + IntToStr(iVer[4]);
  end;
end;

Procedure LoadImage(Filename: aString; Var Error: TSP_ErrorCode);
Var
  Bmp: TPicture;
  Idx: Integer;
  MaxPal: TMaxLogPalette;
  r, g, b: Byte;
  Ext, FirstBytes, OldFilename, tStr: aString;
  FS: TFileStream;
Begin

  // Loads (using arch-specific routines - GraphicEx under Win32, SDL_Image under linux) an image,
  // and fills the supplied pointer and w/h/s vars with info about the image. Use the "FreeImageResource"
  // proc when we're done with the image.

  // FUCKING UGLY HACK ALERT!

  OldFilename := '';
  If FileExists(String(Filename)) Then Begin
    Ext := Lower(aString(ExtractFileExt(String(Filename))));
    FS := TFileStream.Create(String(FileName), fmOpenRead or fmShareDenyNone);
    SetLength(FirstBytes, 8);
    FS.Read(FirstBytes[1], 8);
    FS.Free;
    Ext := '';
    if Copy(FirstBytes, 1, 2) = 'BM' then Ext := '.bmp';
    if FirstBytes = #137'PNG'#13#10#26#10 then Ext := '.png';
    if Copy(FirstBytes, 1, 3) =  'GIF' then Ext := '.gif';
    if Copy(FirstBytes, 1, 2) = #$FF#$D8 then Ext := '.jpg';
    OldFilename := Filename;
    tStr := aString(ExtractFilename(String(Filename)));
    tStr := Copy(tStr, 1, Length(tStr) - Length(Ext));
    Filename := aString(ExtractFilePath(String(Filename))) + tStr + Ext;
    RenameFile(String(OldFilename), String(Filename));
  End;

  If FileExists(String(Filename)) Then Begin
    Bmp := TPicture.Create;
    Try
      Bmp.LoadFromFile(String(Filename));
      Sleep(20);
    Except
      On E: Exception Do Begin
        Bmp.Free;
        Error.Code := SP_ERR_UNSUPPORTED_IMAGE_FORMAT;
      End;
    End;
    If Error.Code = SP_ERR_OK Then Begin
      If Bmp.Bitmap.PixelFormat <> pf8Bit Then
        Error.Code := SP_ERR_INVALID_IMAGE_FORMAT
      Else Begin
        ImgResource := TFastDIB.Create;
        ImgResource.SetSize(Bmp.Width, -Bmp.Height, 8);
        GetPaletteEntries(Bmp.Bitmap.Palette, 0, 256, MaxPal.palPalEntry);
        For Idx := 0 To 255 Do Begin
          r := MaxPal.palPalEntry[Idx].peRed;
          g := MaxPal.palPalEntry[Idx].peGreen;
          b := MaxPal.palPalEntry[Idx].peBlue;
          ImgResource.Colors[Idx].r := r;
          ImgResource.Colors[Idx].g := g;
          ImgResource.Colors[Idx].b := b;
          ImgPalette[Idx].R := r;
          ImgPalette[Idx].G := g;
          ImgPalette[Idx].B := b;
        End;
        ImgResource.UpdateColors;
        BitBlt(ImgResource.hDC, 0, 0, Bmp.Width, Bmp.Height, Bmp.Bitmap.Canvas.Handle, 0, 0, SRCCOPY);
        ImgPtr := pByte(ImgResource.Bits);
        ImgStride := ImgResource.BWidth;
        ImgHeight := Bmp.Height;
        ImgWidth := Bmp.Width;
      End;
      Bmp.Free;
    End Else
      Error.Code := SP_ERR_FILE_MISSING;
  End Else
    Error.Code := SP_ERR_FILE_MISSING;

  If OldFilename <> '' Then
    RenameFile(String(Filename), String(OldFileName));

End;

Procedure FreeImageResource;
Begin

  // Removes an image from memory after loading.

  ImgResource.Free;

End;

end.
