// Copyright (C) 2010 By Paul Dunn
//
// This file is part of the SpecBAS BASIC Interpreter, which is in turn
// part of the SpecOS project.
//
// SpecBAS is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// SpecBAS is distributed in the hope that it will be entertaining,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with SpecBAS.  If not, see <http://www.gnu.org/licenses/>.

unit SP_Editor;

{$IFDEF FPC}
  {$MODE Delphi}
{$ENDIF}
{$INCLUDE SpecBAS.inc}

interface

Uses Math, SP_SysVars, SysUtils, SP_Util, SP_Graphics, SP_BankManager, SP_Tokenise, SP_Errors, SP_Input,
     Classes, SP_InfixToPostFix, SP_Interpret_PostFix, SP_Variables, SP_Sound, SP_Package, SP_FileIO,
     SP_Graphics32;

Procedure SP_CreateEditorWindows;
Procedure SP_HideEditorWindows;
Procedure SP_ShowEditorWindows;
Procedure SP_CycleEditorWindows;
Procedure SP_Decorate_Window(WindowID: Integer; Title: aString; SizeGrip: Boolean);
Procedure SP_DrawBatteryStatus;
Procedure SP_MakeEditorList;
Procedure SP_AutoList;
Procedure SP_EditLoop(Var Error: TSP_ErrorCode);
Procedure SP_GetUserInput;
Procedure SP_EditorHandleMouseDown(X, Y: Integer);
Procedure SP_EditorHandleMouseUp(X, Y: Integer);
Procedure SP_EditorHandleMouseMove(X, Y: Integer);
Procedure SP_EditorHandleMouseWheel(WheelUp: Boolean; X, Y: Integer);
Procedure SP_ResizeCommandWindow(W, H: Integer; Redraw: Boolean);
Procedure SP_MoveCommandWindow(X, Y: Integer; Redraw: Boolean);
Procedure SP_EditorDisplayEditLine;
Procedure SP_EditorPerformEdit(Char: Byte);
Procedure SP_EditorEditLine;
Procedure SP_ScrollInView(LineNum: Integer);
Function  SP_ListWindowShowing: Boolean;
Procedure SP_EditorError(Error: TSP_ErrorCode);
Procedure SP_AddToHistory(Text: aString);
Function  SP_GetInput(Var Error: TSP_ErrorCode): Boolean;
Procedure SP_PerformINPUT(Char: Byte);
Function  SP_INPUTFromFORMAT: aString;
Procedure SP_Execute(Line: aString; Var Error: TSP_ErrorCode);
Function  SP_ShowMenu(Var Options: aString; cX, cY: Integer): Integer;

Var

  TopLine: Integer;
  EditorList: TStringlist;
  EditorMouseDown: Boolean;
  EditorMouseOp: Integer;
  EditorMouseDelay: LongWord;
  EditorMouseY, EditorMouseX, EditorDefaultWindow: Integer;
  EditorMouseControl, EditorFontW, EditorFontH: Integer;
  EditorClickTime, EditorClickX, EditorClickY: LongWord;
  EditorDRPOSX, EditorDRPOSY, EditorPRPOSY, EditorPRPOSX: Integer;
  EditorOVER: Integer;
  EditorMouseStatus: Boolean;
  EditorMarks: Array[0..9] of LongWord;
  EditorHistory: Array of aString;
  HistoryPos: Integer;
  EditorFPS: aFloat;
  EditorFRAME_MS: Integer;
  EditorOutSet: Boolean;
  EdSc: aString;

  UsrCWX, UsrCWY, UsrCWW, UsrCWH: Integer;

implementation

Uses SP_Main;

Function SP_SetEditorFont: Integer;
Var
  Err: TSP_ErrorCode;
Begin

  Result := FONTBANKID;
  SP_SetSystemFont(EDITORFONT, Err);
  EditorFontW := Trunc(FONTWIDTH * EDFONTSCALEX);
  EditorFontH := Trunc(FONTHEIGHT * EDFONTSCALEY);
  EdSc := #25 + aFloatToString(EDFONTSCALEX) + aFloatToString(EDFONTSCALEY);

End;

Procedure SP_CreateEditorWindows;
Var
  CWin: pSP_Window_Info;
  Err: TSP_ErrorCode;
  Font, Idx: Integer;
Begin

  SP_DisableDisplay;

  If CCOMMANDWINDOW or CLISTWINDOW Then SP_HideEditorWindows;

  EditorDefaultWindow := SCREENID;
  EditorDRPOSX := DRPOSX;
  EditorDRPOSY := DRPOSY;
  EditorPRPOSX := PRPOSX;
  EditorPRPOSY := PRPOSY;
  EditorOVER := COVER;
  EditorFPS := FPS;
  EditorFRAME_MS := FRAME_MS;

  COVER := 0;
  T_OVER := COVER;
  FPS := 50;
  FRAME_MS := 20;

  SP_SetDrawingWindow(SCREENBANK, 0);
  Font := SP_SetEditorFont;

  If LISTWINDOW = -1 Then Begin

    // Set up the listing and direct command windows for the first time

    LISTWINDOW := SP_Add_Window(SCREENBANK, 4, 4, DISPLAYWIDTH - 8, DISPLAYHEIGHT - ((EditorFontH * 4) + 4) -14, -1, 8, 0, Err);
    COMMANDWINDOW := SP_Add_Window(SCREENBANK, 4, DISPLAYHEIGHT - ((EditorFontH * 4) + 9), DISPLAYWIDTH - 8, (EditorFontH * 4) + 4, -1, 8, 0, Err);

  End Else Begin

    // Restore the windows from previous values.

    If LISTWINW > DISPLAYWIDTH Then LISTWINW := DISPLAYWIDTH - 8;
    If LISTWINX > DISPLAYWIDTH Then LISTWINX := (DISPLAYWIDTH -4) - LISTWINW;
    If LISTWINH > DISPLAYHEIGHT Then LISTWINH := DISPLAYHEIGHT - 8;
    If LISTWINY > DISPLAYHEIGHT Then LISTWINY := (DISPLAYHEIGHT -4) - LISTWINH;

    If UsrCWW > DISPLAYWIDTH Then UsrCWW := DISPLAYWIDTH - 8;
    If UsrCWX > DISPLAYWIDTH Then UsrCWX := (DISPLAYWIDTH -4) - UsrCWW;
    If UsrCWH > DISPLAYHEIGHT Then UsrCWH := DISPLAYHEIGHT - 8;
    If UsrCWY > DISPLAYHEIGHT Then UsrCWY := (DISPLAYHEIGHT -4) - UsrCWH;

    LISTWINDOW := SP_Add_Window(SCREENBANK, LISTWINX, LISTWINY, LISTWINW, LISTWINH, -1, 8, 0, Err);
    COMMANDWINDOW := SP_Add_Window(SCREENBANK, UsrCWX, UsrCWY, UsrCWW, UsrCWH, -1, 8, 0, Err);

  End;

  SP_GetWindowDetails(LISTWINDOW, CWin, SP_BankList[0], Err);
  For Idx := 0 To 255 Do CWin^.Palette[Idx] := DefaultPalette[Idx];
  CWin^.Visible := False;
  CWin^.Over := 0;

  SP_GetWindowDetails(COMMANDWINDOW, CWin, SP_BankList[0], Err);
  For Idx := 0 To 255 Do CWin^.Palette[Idx] := DefaultPalette[Idx];

  COMMANDWINX := CWin^.Left;
  COMMANDWINY := CWin^.Top;
  COMMANDWINW := CWin^.Width;
  COMMANDWINH := CWin^.Height;
  COMMANDSIZETLX := CWin^.Width - (EditorFontW + 6);
  COMMANDSIZETLY := CWin^.Height - (EditorFontH + 6);
  COMMANDSIZEBRX := CWin^.Width - 6;
  COMMANDSIZEBRY := CWin^.Height - 6;
  CWin^.Over := 0;
  CWin^.Visible := False;

  UsrCWX := COMMANDWINX;
  UsrCWY := COMMANDWINY;
  UsrCWW := COMMANDWINW;
  UsrCWH := COMMANDWINH;

  SP_Decorate_Window(COMMANDWINDOW, 'Direct command', True);
  SP_Decorate_Window(LISTWINDOW, 'Program listing - ' + PROGNAME, True);

  CCOMMANDWINDOW := True;
  CLISTWINDOW := True;

  EditorMouseStatus := MOUSEVISIBLE;
  MOUSEVISIBLE := True;

  SP_SetSystemFont(Font, Err);
  SP_MakeEditorList;
  SP_EnableDisplay;

End;

Procedure SP_Decorate_Window(WindowID: Integer; Title: aString; SizeGrip: Boolean);
Var
  Win: pSP_Window_Info;
  Err: TSP_ErrorCode;
  Idx, Font, Window: Integer;
  tTitleBar: aString;
Begin

  Window := SCREENID;
  Font := SP_SetEditorFont;

  SP_GetWindowDetails(WindowID, Win, SP_BankList[0], Err);
  SP_SetDrawingWindow(SCREENBANK, WindowID);
  T_INK := 0;
  T_OVER := 0;

  SP_FillRect(0, EditorFontH + 2, Win^.Width, 4, 7); // 4 pixel border at the top
  SP_FillRect(0, EditorFontH + 6, 4, Win^.Height, 7); // And to the left
  SP_FillRect(4, Win^.Height - (EditorFontH + 6), Win^.Width, EditorFontW + 6, 7);
  SP_FillRect(Win^.Width - (EditorFontW + 7), EditorFontH + 6, EditorFontW + 7, Win^.Height - (EditorFontH + 6), 7);
  SP_FillRect(0, 0, Win^.Width, EditorFontH +2, 0);

  SP_DrawRectangle(0, 0, Win^.Width, Win^.Height);
  SP_TextOut(-1, 1, 1, EdSc + Title, 15, 0, True);
  SP_TextOut(-1, Win^.Width - ((EditorFontW * 6) + 2), 1, EdSc + #16#10#0#0#0#17#0#0#0#0#255#16#14#0#0#0#17#10#0#0#0#255#16#12#0#0#0#17#14#0#0#0#255#16#13#0#0#0#17#12#0#0#0#255#16#0#0#0#0#17#13#0#0#0#255, 15, 0, True);

  If SizeGrip Then
    SP_TextOut(-1, Win^.Width -(EditorFontW + 6), Win^.Height - (EditorFontH + 6), EdSc + #250, 0, 7, True);

  SP_SetDrawingWindow(SCREENBANK, Window);
  SP_SetSystemFont(Font, Err);

End;

Procedure SP_DrawBatteryStatus;
Var
  Idx, X, Y, BattW, pW, PixW, Font, Window: Integer;
  PixPtr: pByte;
  Win: pSP_Window_Info;
  Err: TSP_ErrorCode;
Begin

  {$IFDEF PANDORA}
  BATTLEVEL := StrToInt(ReadLinuxFile('/sys/class/power_supply/bq27500-0/capacity'));
  {$ELSE}
  BATTLEVEL := 100;
  {$ENDIF}

  // Draws the battery status in the current window. Assumes the current window is a "system" window, and has a stripe.
  // First, find the stripe so we know what to do!

  Window := SCREENID;
  Font := SP_SetEditorFont;

  SP_GetWindowDetails(COMMANDWINDOW, Win, SP_BankList[0], Err);
  SP_SetDrawingWindow(SCREENBANK, COMMANDWINDOW);

  T_INK := 0;
  T_OVER := 0;

  X := Win^.Width - ((EditorFontW * 6) + 2);
  Y := 1 + EditorFontH;

  // Now calculate how much we need to remove for the battery meter.
  // There are four stripes.

  BattW := Round((((EditorFontW * 4) -1)) * ((100-BATTLEVEL)/100));
  SP_TextOut(-1, Win^.Width - ((EditorFontW * 6) + 2), 1, EdSc + #16#10#0#0#0#17#0#0#0#0#255#16#14#0#0#0#17#10#0#0#0#255#16#12#0#0#0#17#14#0#0#0#255#16#13#0#0#0#17#12#0#0#0#255#16#0#0#0#0#17#13#0#0#0#255, 15, 0, True);

  For Idx := 2 To EditorFontH -1 Do Begin
    PixPtr := @SP_BankList[0]^.Memory[Win^.Offset + LongWord(X + (Y * Win^.Width) - (Idx * Win^.Width) + Idx + (EditorFontW * 4) -2)];
    PixW := BattW;
    pW := 0;
    While PixW > 0 Do Begin
      If (pW Mod EditorFontW <> 0) And (pW Mod EditorFontW <> EditorFontW -1) Then
        PixPtr^ := PixPtr^ - 8;
      Dec(PixW);
      Dec(PixPtr);
      Inc(pW);
    End;
  End;

  SP_SetDrawingWindow(SCREENBANK, Window);
  SP_SetSystemFont(Font, Err);

End;

Function SP_GetLength(Const Text: aString): Integer;
Var
  pChr: pByte;
  Idx, Len: Integer;
Begin

  Result := 0;
  pChr := @Text[1];
  Idx := 1;
  Len := Length(Text);
  While Idx <= Len Do Begin
    If pChr^ >= 32 Then
      Inc(Result);
    Inc(pChr);
    Inc(Idx);
  End;

End;

Procedure SP_MakeEditorList;
Var
  Err: TSP_ErrorCode;
  Win: pSP_Window_Info;
  CodeLine, Text: aString;
  FirstLine, InString, InREM, FirstREM: Boolean;
  Idx, Idx2, Idx3, Ink, LineNum, IndentLevel, WinW, ProgLen,
  Idx4, MaxLen, LineNumSize, WinCW, cPos, Len, Font, tIdx: Integer;
Begin

  // Makes the line list for the editor. Splits statements, wraps lines etc. Detects REM lines
  // and formats accordingly.

  cPos := -1;
  IndentLevel := 7;
  LineNum := 0;
  FirstREM := False;

  SP_GetWindowDetails(LISTWINDOW, Win, SP_BankList[0], Err);

  If (Err.Code <> SP_ERR_BANK_NOT_FOUND) And (Err.Code <> SP_ERR_WINDOW_NOT_FOUND) Then Begin

    EditorList.Clear;
    Font := SP_SetEditorFont;
    WinW := Win^.Width - (14 + EditorFontW *2);
    WinCW := WinW Div EditorFontW;

    If INCLUDEFROM = -1 Then
      ProgLen := Length(SP_Program) -1
    Else
      ProgLen := INCLUDEFROM -1;

    For Idx := 0 To ProgLen Do Begin

      // Get the plain text line, and skip the line number.

      CodeLine := SP_Detokenise(SP_Program[Idx], cPos, True, False);
      Idx2 := 1; While CodeLine[Idx2] in ['0'..'9', ' '] Do Inc(Idx2);
      LineNumSize := Idx2 -1;

      InREM := False;
      FirstLine := True;
      Ink := 0;

      While Length(CodeLine) > 0 Do Begin

        Len := 0;
        Idx2 := 1;
        InString := False;
        If Not FirstLine Then Begin
          CodeLine := aString(#16 + LongWordToString(Ink)) + CodeLine;
          LineNumSize := 0;
        End;
        While Idx2 < Length(CodeLine) Do Begin
          If CodeLine[Idx2] = #16 Then Begin
            Ink := pLongWord(@CodeLine[Idx2 +1])^;
            Inc(Idx2, 1 + SizeOf(LongWord));
            If Idx2 >= Length(CodeLine) Then Break;
          End;
          If CodeLine[Idx2] >= ' ' Then Inc(Len);
          If CodeLine[Idx2] = '"' Then
            If InREM Then
              InString := False
            Else
              InString := Not InString;
          If (Not InString And (Copy(CodeLine, Idx2, 3) = 'REM') and (Copy(CodeLine, Idx2 + 3, 1) <= ' ') and (Copy(CodeLine, Idx2 -1, 1) <= ' ')) or InREM Then Begin
            Ink := 8;
            If SPLITREMS Then Begin
              If Not InREM Then
                Idx2 := Max(Idx2 + 5, Min(Length(CodeLine), Idx2 + WinCW - (IndentLevel - 5)))
              Else
                Idx2 := Max(Idx2 + 5, Min(Length(CodeLine), Idx2 + WinCW - IndentLevel - 3));
              If Idx2 < Length(CodeLine) Then Begin
                While (Idx2 > 1) And (CodeLine[Idx2] in ['a'..'z', 'A'..'Z', '0'..'9']) Do
                  Dec(Idx2);
                If Not InREM Then Begin
                  FirstREM := True;
                  InREM := True;
                End Else
                  FirstREM := False;
              End Else
                If Not InREM Then Begin
                  FirstREM := True;
                  InREM := True;
                End Else
                  FirstREM := False;
            End Else
              Idx2 := Length(CodeLine);
            Break;
          End;
          If (CodeLine[Idx2] = ':') Then Begin
            If Not (InString Or InREM) and SPLITSTATEMENTS Then
              Break;
            tIdx := Idx2 +1;
            While CodeLine[tIdx] <= ' ' Do Inc(tIdx);
            If SPLITREMS And Not InString And Not InREM And (Copy(CodeLine, tIdx, 3) = 'REM') Then
              Break;
          End;
          If CodeLine[Idx2] = #13 Then
            If SPLITSTATEMENTS Then
              Break
            Else Begin
              CodeLine := Copy(CodeLine, 1, Idx2 -1) + Copy(CodeLine, Idx2 +1, Length(CodeLine));
              Dec(Idx2);
            End;
          If EDITORWRAP Then
            If Len = WinCW -6 + LineNumSize Then
              Break;
          Inc(Idx2);
        End;

        Text := Copy(CodeLine, 1, Idx2);
        If FirstLine Then Begin
          While (Text <> '') And (Text[1] = ' ') Do
            Text := Copy(Text, 2, 99999);
        End Else Begin
          Idx3 := 1;
          While (Idx3 <= Length(Text)) And (Text[Idx3] < ' ') Do
            Inc(Idx3);
          While (Idx3 <= Length(Text)) And (Text[Idx3] = ' ') Do
            Text := Copy(Text, 1, Idx3 -1) + Copy(Text, Idx3 +1, 99999);
          If Idx3 > Length(Text) Then Text := '';
          If InREM And Not FirstREM Then Text := '    ' + Text;
        End;

        If Text <> '' Then Begin
          If FirstLine or Not InREM Then Begin
            If Text[1] in ['0'..'9'] Then Begin
              Idx3 := 1;
              LineNum := 0;
              While Text[Idx3] in ['0'..'9'] Do Begin
                LineNum := (LineNum * 10) + (Ord(Text[Idx3])-48);
                Inc(Idx3);
              End;
            End Else
              Idx3 := 0;
          End Else
            Idx3 := 0;

          While Idx3 < IndentLevel Do Begin
            Text := ' ' + Text;
            Inc(Idx3);
          End;

          If InREM And Not FirstLine Then
            Text := #16#8#0#0#0 + Text;
          EditorList.Add(String(Text));
          EditorList.Objects[EditorList.Count -1] := TObject(LineNum);
          FirstLine := False;
          Idx4 := SP_GetLength(Text);
          If Idx4 > MaxLen Then MaxLen := Idx4;
        End;

        CodeLine := Copy(CodeLine, Idx2 +1, 999999);
        While (Idx2 < Length(CodeLine)) and (CodeLine[Idx2] <= ' ') Do Inc(Idx2);

      End;

    End;

    SP_SetSystemFont(Font, Err);

  End;

End;

Procedure SP_AutoList;
Var
  Idx, Idx2, Idx3, Idx4, cPos, X, Y, Font, Window, MaxLen, IndentLevel, WinCW, Ink, LineNumSize,
  PgSize, WinX, WinY, WinH, WinW, Paper, Len, CurIndent: Integer;
  CodeLine, Statement, Text: aString;
  Win: pSP_Window_Info;
  Err: TSP_ErrorCode;
  Chr: pAnsiChar;
  InString, InREM, FirstLine: Boolean;
  LineNum: LongWord;
Begin

  Window := SCREENID;
  SP_DisableDisplay;

  Font := SP_SetEditorFont;
  SP_SetDrawingWindow(SCREENBANK, LISTWINDOW);
  SP_GetWindowDetails(LISTWINDOW, Win, SP_BankList[0], Err);

  WinX := 4;
  WinY := EditorFontH + 6;
  WinH := Win^.Height - WinY - EditorFontH - 4 - 4 - 2;
  WinW := Win^.Width - (14 + EditorFontW); // two borders of 4 pixels, one character, then 4 pixels, with 2 pixel black border.

  If Not CLISTWINDOW Then Exit;

  // List the program from the topindex down

  T_INK := 0;
  T_OVER := 0;

  SP_FillRect(1, EditorFontH + 2, Win^.Width -2, Win^.Height - EditorFontH -2, 7);

  LISTLINECOUNT := EditorList.Count;
  MaxLen := 0;
  For Idx := 0 To EditorList.Count -1 Do Begin
    LISTCOLCOUNT := 0;
    Idx3 := Length(EditorList[Idx]);
    While EditorList[Idx][Idx3] <= ' ' Do
      Dec(Idx3);
    For Idx2 := 1 To Idx3 Do
      If EditorList[Idx][Idx2] >= ' ' Then
        Inc(LISTCOLCOUNT);
    If LISTCOLCOUNT > MaxLen Then
      MaxLen := LISTCOLCOUNT;
  End;
  LISTCOLCOUNT := MaxLen;
  LISTVISLINES := (WinH Div EditorFontH);
  LISTVISCOLS := (WinW Div EditorFontW);
  LISTWINX := Win^.Left;
  LISTWINY := Win^.Top;
  LISTWINW := Win^.Width;
  LISTWINH := Win^.Height;

  LISTSIZETLX := Win^.Width - (EditorFontW + 6);
  LISTSIZETLY := Win^.Height - (EditorFontH + 6);
  LISTSIZEBRX := Win^.Width - 6;
  LISTSIZEBRY := Win^.Height - 6;

  SP_TextOut(-1, LISTSIZETLX, LISTSIZETLY, EdSc + #250, 0, 7, True);

  If LISTTOPINDEX + LISTVISLINES > EditorList.Count Then
    LISTTOPINDEX := Max(0, EditorList.Count - (LISTVISLINES -1));

  Y := EditorFontH + 6;
  Idx := Round(LISTTOPINDEX);
  SP_DrawRectangle(0, 0, Win^.Width, Win^.Height);

  If EditorList.Count > 0 Then

    While Idx < LISTTOPINDEX + LISTVISLINES Do Begin

      // Search the line for "real" characters
      CodeLine := aString(EditorList[Idx]) + #$FF;
      Idx2 := Round(LISTLEFTINDEX) +1;
      Idx3 := 1; Idx4 := Length(CodeLine);
      Len := 0;
      Chr := @CodeLine[1];
      T_INK := 0;
      While Len < Idx2 Do Begin
        If Chr^ >= ' ' Then Inc(Len);
        If Chr^ = #16 Then Begin
          T_INK := pLongWord(LongWord(Chr) +1)^;
          Inc(Chr, SizeOf(LongWord));
        End;
        Inc(Chr);
        Inc(Idx3);
        If Idx3 > Idx4 Then Break;
      End;

      If Idx3 <= Idx4 Then Begin
        Text := '';
        Chr := @CodeLine[Idx3];
        Idx3 := Min(LISTVISCOLS + Round(LISTLEFTINDEX), Length(EditorList[Idx]));
        While Idx2 <= Idx3 Do Begin
          If Chr^ < #$FF Then Begin
            Text := Text + Chr^;
            If Chr^ >= ' ' Then Inc(Idx2);
            Inc(Chr);
          End Else
            Break;
        End;
        If (Integer(EditorList.Objects[Idx]) = PROGLINE) Then Begin
          Paper := 5;
          SP_FillRect(WinX, Y, WinW, EditorFontH, 5);
        End Else
          Paper := 7;
        SP_TextOut(-1, WinX, Y, EdSc + Text, T_INK, Paper, True);
      End Else
        If (Integer(EditorList.Objects[Idx]) = PROGLINE) Then Begin
          SP_FillRect(WinX, Y, WinW, EditorFontH, 5);
        End;

      Inc(Idx);
      If Idx = EditorList.Count Then Break;
      Inc(Y, EditorFontH);

    End;

  // Now draw the scroll bars - vertical

  // Up arrow

  LISTVSCUPTLX := LISTSIZETLX;
  LISTVSCUPTLY := WinY;
  LISTVSCUPBRX := LISTVSCUPTLX + EditorFontW;
  LISTVSCUPBRY := LISTVSCUPTLY + EditorFontH;

  If LISTTOPINDEX > 0 Then Begin
    SP_TextOut(-1, LISTVSCUPTLX, LISTVSCUPTLY, EdSc + #251, 0, 7, True);
    LISTVSCUPENABLED := True;
  End Else Begin
    SP_TextOut(-1, LISTVSCUPTLX, LISTVSCUPTLY, EdSc + #251, 8, 7, True);
    LISTVSCUPENABLED := False;
  End;

  // Down arrow

  LISTVSCDNTLX := LISTVSCUPTLX;
  LISTVSCDNTLY := WinY + WinH - EditorFontH;
  LISTVSCDNBRX := LISTVSCDNTLX + EditorFontW;
  LISTVSCDNBRY := LISTVSCDNTLY + EditorFontH;

  If LISTTOPINDEX + LISTVISLINES < LISTLINECOUNT Then Begin
    SP_TextOut(-1, LISTVSCDNTLX, LISTVSCDNTLY, EdSc + #252, 0, 7, True);
    LISTVSCDNENABLED := True;
  End Else Begin
    SP_TextOut(-1, LISTVSCDNTLX, LISTVSCDNTLY, EdSc + #252, 8, 7, True);
    LISTVSCDNENABLED := False;
  End;

  // Vertical Track - should be the same height as the "page", with two characters removed

  LISTVSCTRKH := LISTVSCDNTLY - LISTVSCUPBRY;

  If LISTLINECOUNT > 0 Then Begin
    If LISTVISLINES < LISTLINECOUNT Then
      PgSize := Round(LISTVSCTRKH * (LISTVISLINES/LISTLINECOUNT))
    Else
      PgSize := LISTVSCTRKH;
  End Else
    PgSize := LISTVSCTRKH;

  LISTVTHS := PgSize;
  If LISTLINECOUNT > 0 Then
    LISTVTHP := Min(LISTVSCUPBRY + Round(LISTTOPINDEX * (LISTVSCTRKH/LISTLINECOUNT)), LISTVSCDNTLY - LISTVTHS)
  Else
    LISTVTHP := LISTVSCUPBRY;

  LISTVSTRTLX := LISTVSCUPTLX;
  LISTVSTRTLY := LISTVSCUPTLY + EditorFontH;
  LISTVSTRBRX := LISTVSTRTLX + EditorFontW;
  LISTVSTRBRY := LISTVSCUPBRY + LISTVSCTRKH;

  If (PgSize > 0) And (PgSize < LISTVSCTRKH) Then Begin
    SP_FillRect(LISTVSTRTLX, LISTVSTRTLY, LISTVSTRBRX - LISTVSTRTLX, LISTVSTRBRY - LISTVSTRTLY, 8);
    SP_FillRect(Win^.Width - (EditorFontW + 6), LISTVTHP, EditorFontW, PgSize, 0);
    LISTVSTRENABLED := True;
  End Else Begin
    SP_FillRect(LISTVSTRTLX, LISTVSTRTLY, LISTVSTRBRX - LISTVSTRTLX, LISTVSTRBRY - LISTVSTRTLY, 8);
    LISTVSTRENABLED := False;
  End;

  // Draw the horizontal scrollbar

  // Left Arrow

  LISTHSCUPTLX := WinX;
  LISTHSCUPTLY := WinY + WinH + 4;
  LISTHSCUPBRX := LISTHSCUPTLX + EditorFontW;
  LISTHSCUPBRY := LISTHSCUPTLY + EditorFontH;

  If LISTLEFTINDEX > 0 Then Begin
    SP_TextOut(-1, LISTHSCUPTLX, LISTHSCUPTLY, EdSc + #254, 0, 7, True);
    LISTHSCUPENABLED := True;
  End Else Begin
    SP_TextOut(-1, LISTHSCUPTLX, LISTHSCUPTLY, EdSc + #254, 8, 7, True);
    LISTHSCUPENABLED := False;
  End;

  // Right Arrow

  LISTHSCDNTLX := LISTSIZETLX - EditorFontW - 4;
  LISTHSCDNTLY := LISTSIZETLY;
  LISTHSCDNBRX := LISTHSCDNTLX + EditorFontW;
  LISTHSCDNBRY := LISTHSCDNTLY + EditorFontH;

  If LISTLEFTINDEX + LISTVISCOLS < MaxLen Then Begin
    SP_TextOut(-1, LISTHSCDNTLX, LISTHSCDNTLY, EdSc + #253, 0, 7, True);
    LISTHSCDNENABLED := True;
  End Else Begin
    SP_TextOut(-1, LISTHSCDNTLX, LISTHSCDNTLY, EdSc + #253, 8, 7, True);
    LISTHSCDNENABLED := False;
  End;

  // Horizontal track

  LISTHSTRTLX := LISTHSCUPTLX + EditorFontW;
  LISTHSTRTLY := LISTHSCUPTLY;
  LISTHSTRBRX := LISTHSCDNTLX;
  LISTHSTRBRY := LISTHSTRTLY + EditorFontH;

  If MaxLen > LISTVISCOLS Then Begin
    LISTHTHS := Round(LISTVISCOLS * ((LISTHSTRBRX - LISTHSTRTLX)/MaxLen));
    LISTHTHP := LISTHSTRTLX + Round(LISTLEFTINDEX * ((LISTHSTRBRX - LISTHSTRTLX)/MaxLen));
    SP_FillRect(LISTHSTRTLX, LISTHSTRTLY, LISTHSTRBRX - LISTHSTRTLX, EditorFontH, 8);
    SP_FillRect(LISTHTHP, LISTHSTRTLY, LISTHTHS, EditorFontH, 0);
    LISTHSTRENABLED := True;
  End Else Begin
    LISTHTHP := LISTHSTRTLX;
    LISTHTHS := LISTHSTRBRX - LISTHSTRTLX;
    SP_FillRect(LISTHSTRTLX, LISTHSTRTLY, LISTHSTRBRX - LISTHSTRTLX, EditorFontH, 8);
    LISTHSTRENABLED := False;
  End;

  // And reset the system fonts and current window

  SP_SetSystemFont(Font, Err);
  SP_SetDrawingWindow(SCREENBANK, Window);
  SP_EnableDisplay;

End;

Procedure SP_ScrollInView(LineNum: Integer);
Var
  Idx: Integer;
Begin

  // First, find the line.

  SP_MakeEditorList;

  Idx := 0;
  While Idx < EditorList.Count Do
    If Integer(EditorList.Objects[Idx]) >= LineNum Then
      Break
    Else
      Inc(Idx);

  If Idx < EditorList.Count Then Begin
    If Idx < LISTTOPINDEX Then
      LISTTOPINDEX := Idx
    Else
      If Idx >= (LISTTOPINDEX + LISTVISLINES)-1 Then
        LISTTOPINDEX := Idx - LISTVISLINES +2;
  End;

  Inc(Idx);
  If Idx < EditorList.Count Then Begin
    While (Idx < EditorList.Count) And (Integer(EditorList.Objects[Idx]) = PROGLINE) Do Begin
      If Idx > (LISTTOPINDEX + LISTVISLINES) -1 Then
        LISTTOPINDEX := LISTTOPINDEX + 1;
      Inc(Idx);
    End;
  End;

End;

Procedure SP_HideEditorWindows;
Var
  Err: TSP_ErrorCode;
  Idx: Integer;
Begin

  SP_DeleteWindow(LISTWINDOW, Err);
  SP_DeleteWindow(COMMANDWINDOW, Err);

  CCOMMANDWINDOW := False;
  CLISTWINDOW := False;

  FPS := EditorFPS;
  FRAME_MS := EditorFRAME_MS;

  CINVERSE := EDINVERSE;
  T_INVERSE := CINVERSE;

  SP_SetDrawingWindow(SCREENBANK, EditorDefaultWindow);
  SP_NeedDisplayUpdate := True;
  CauseUpdate := True;
  MOUSEVISIBLE := EditorMouseStatus;

End;

Procedure SP_Interpreter(Var Tokens: paString; Var Position: Integer; Var Error: TSP_ErrorCode; PreParseErrorCode: Integer);
Var
  CurLine, ProgLen: Integer;
Begin

  SP_Interpret(Tokens, Error.Position, Error);
  If Error.ReturnType = SP_JUMP Then
    If PreParseErrorCode <> SP_ERR_OK Then Begin
      Error.Code := PreParseErrorCode;
      Error.ReturnType := 0;
    End;

  If INCLUDEFROM > -1 Then Begin
    If NXTLINE >= INCLUDEFROM Then NXTLINE := -1;
  End Else
    If NXTLINE >= SP_Program_Count Then NXTLINE := -1;

  While NXTLINE <> -1 Do Begin

    If NXTLINE = -2 Then Begin
      CurLine := -1;
      Tokens := @COMMAND_TOKENS;
      If NXTSTATEMENT = -1 Then Begin Dec(Error.Statement); Exit; End;
      If Byte(Tokens^[NXTSTATEMENT]) = SP_TERMINAL Then
        Dec(Error.Statement);
    End Else Begin
      CurLine := NXTLINE;
      Tokens := @SP_Program[CurLine];
    End;
    If NXTSTATEMENT <> -1 Then
      Error.Position := NXTSTATEMENT
    Else Begin
      Error.Statement := 1;
      Error.Position := SP_FindStatement(Tokens, 1);
    End;

    NXTSTATEMENT := -1;
    Inc(NXTLINE);
    If NXTLINE <> 0 Then Begin
      Error.Line := CurLine;
      SP_StackPtr := SP_StackStart;
      SP_Interpret(Tokens, Error.Position, Error);
    End;

    If NXTLINE = SP_Program_Count Then NXTLINE := -1;

    If Error.Code <> SP_ERR_OK Then Begin
      CONTLINE := Error.Line;
      CONTSTATEMENT := Error.Statement +1;
      NXTLINE := -1;
    End Else Begin
      If NXTLINE <> -1 Then Begin
        If NXTLINE >= SP_Program_Count Then
          NXTLINE := -1;
      End;

    End;

  End;

End;

Procedure SP_Execute(Line: aString; Var Error: TSP_ErrorCode);
Var
  Tokens: paString;
  aSave: Boolean;
Begin

  aSave := AUTOSAVE;
  AUTOSAVE := False;

  Error.Line := -1;
  Error.Statement := 1;
  Error.Position := 1;
  Line := SP_TokeniseLine(Line, False, False) + #255#255#255#255;
  SP_Convert_ToPostFix(Line, Error.Position, Error);
  Tokens := @Line;
  Error.Position := SP_FindStatement(@Line, 1);
  Error.Code := SP_ERR_OK;
  COMMAND_TOKENS := Line;
  NXTSTATEMENT := -1;
  NXTLINE := -1;
  SP_StackPtr := SP_StackStart;
  SP_PreParse(False, Error);
  PROGSTATE := SP_PR_RUN;
  SP_Interpreter(Tokens, Error.Position, Error, Error.Code);

  AUTOSAVE := aSave;

End;

Procedure SP_ShowEditorWindows;
Var
  Error: TSP_ErrorCode;
Begin

  SP_SetWindowVisible(LISTWINDOW, True, Error);
  SP_SetWindowVisible(COMMANDWINDOW, True, Error);

End;

Procedure SP_CycleEditorWindows;
Var
  Idx: Integer;
  Error: TSP_ErrorCode;
Begin

  If (KEYSTATE[K_SHIFT] = 1) or (EDITORHIDDEN <> 0) Then Begin
    If SP_WindowVisible(LISTWINDOW, Error) Then EDITORHIDDEN := 1 Else EDITORHIDDEN := 0;
    If SP_WindowVisible(COMMANDWINDOW, Error) Then EDITORHIDDEN := EDITORHIDDEN + 2;
    SP_SetWindowVisible(LISTWINDOW, False, Error);
    SP_SetWindowVisible(COMMANDWINDOW, False, Error);
    MOUSEVISIBLE := EditorMouseStatus;
    LASTKEY := 0;
    Repeat
      CB_YIELD;
    Until Not (LASTKEY in [0, K_SHIFT]);
    LASTKEY := 0;
    If EDITORHIDDEN And 1 = 1 Then
      SP_SetWindowVisible(LISTWINDOW, True, Error);
    If EDITORHIDDEN And 2 = 2 Then
      SP_SetWindowVisible(COMMANDWINDOW, True, Error);
    EDITORHIDDEN := 0;
    EditorMouseStatus := MOUSEVISIBLE;
    MOUSEVISIBLE := True;
  End Else
    If SP_WindowVisible(LISTWINDOW, Error) Then // List window showing; just hide list window
      SP_SetWindowVisible(LISTWINDOW, False, Error)
    Else
      SP_SetWindowVisible(LISTWINDOW, True, Error);

End;

Procedure SP_EditLoop(Var Error: TSP_ErrorCode);
Var
  aSave: Boolean;
  TokensStr, Expr: aString;
  Tokens: paString;
  PreParseErrorCode, PreParseErrorLine, PreParseErrorStatement, CurLine, Idx, LocalFlashState: Integer;
  pInfo: pSP_iInfo;
  Info: TSP_iInfo;
Label
  RunTimeExit;
Begin

  Info.Error := @Error;
  pInfo := @Info;

  {$IFDEF PANDORA}
  BATTLEVEL := StrToInt(ReadLinuxFile('/sys/class/power_supply/bq27500-0/capacity'));
  {$ELSE}
  BATTLEVEL := 100;
  {$ENDIF}

  EditorMouseDown := False;
  EditorMouseControl := 0;
  EditorList := TStringlist.Create;

  SPLITSTATEMENTS := TRUE;
  EDITORWRAP := FALSE;
  SPLITREMS := FALSE;

  // Execute the startup code

  aSave := AUTOSAVE;
  AUTOSAVE := False;
  If SP_FileExists('s:startup-sequence') Then Begin
    SP_Execute('LOAD "s:startup-sequence": RUN', Error);
    // Clear any errors, as we just ignore them.
    Error.Code := SP_ERR_OK;
  End;

  // Startup.
  // Check if the user supplied any parameters. If so, use the first as the mandatory filename,
  // and the second as the optional line number to start from.

  If ParamCount > 0 Then Begin
    NXTLINE := -1;

    SP_LoadProgram(aString(ParamStr(1)), False, True, nil, Error);
    If Error.Code = SP_ERR_OK Then Begin
      If ParamCount > 1 Then
        CurLine := StrToIntDef(ParamStr(2), -1)
      Else
        CurLine := NXTLINE;
      If CurLine = -1 Then CurLine := 0;
      SP_PreParse(True, Error);
      SetExceptionMask([exInvalidOp, exDenormalized, exZeroDivide, exOverflow, exUnderflow, exPrecision]);

      SP_Execute('RUN '+IntToString(CurLine), Error);
    End;
    SP_Interpret_QUIT(pInfo);
    Exit;
  End Else Begin

    // Now load the auto-save if it exists, and set up for a NEW error message.

    If aSave And (SP_FileExists('s:autosave')) Then Begin
      SP_LoadProgram('s:autosave', False, False, nil, Error);
    End Else Begin
      SP_Program_Count := 0;
      Error.Line := -2;
      Error.Statement := 0;
      Error.Code := -1;
      NXTLINE := -1;
      NXTSTATEMENT := -1;
      Error.Position := 1;
      PROGNAME := 'New program';
      FILENAMED := False;
      SP_PreParse(True, Error);
    End;

    Error.Line := -2;
    Error.Statement := 0;
    Error.Code := -1;
    NXTLINE := -1;
    Error.Position := 1;

  End;

  AUTOSAVE := aSave;

  Repeat

    SP_Reset_Temp_Colours;

    If Error.Code <> SP_ERR_NO_ERROR Then
      If Not EDITERROR Then Begin
        PROGSTATE := SP_PR_STOP;
        SP_EditorError(Error);
        PreparseErrorCode := SP_ERR_OK;
        COVER := 0;
        T_OVER := 0;
      End;

    If LASTKEY < 32 Then LASTKEY := 0;

    If (Not CCOMMANDWINDOW) or (COMMANDWINDOW = -1) Then Begin
      SP_CreateEditorWindows;
      Error.Code := SP_ERR_OK;
    End;

    SP_SetWindowVisible(COMMANDWINDOW, True, Error);

    If SHOWLIST Then Begin
      SP_ShowEditorWindows;
      SHOWLIST := False;
    End;

    SP_AutoList;
    SP_GetUserInput;
    If (StripLeadingSpaces(EDITLINE) <> '') And (KeyState[K_SHIFT] = 0) Then Begin

      TokensStr := SP_TokeniseLine(EDITLINE, False, True) + SP_TERMINAL_SEQUENCE;
      SP_Convert_ToPostFix(TokensStr, Error.Position, Error);
      EDITLINE := SP_DeTokenise(TokensStr, Error.Position, False, False);
      If Error.Code = SP_ERR_OK Then Begin
        EDITERROR := False;
        If TokensStr[1] = aChar(SP_LINE_NUM) Then Begin
          SP_DeleteIncludes;
          EDITLINE := '';
          If Not SP_WindowVisible(LISTWINDOW, Error) Then
            SP_SetWindowVisible(LISTWINDOW, True, Error);
          DoAutoSave(Error);
          PROGLINE := dLongWord(@TokensStr[2]);
          ProcListAvailable := False;
          SP_MakeEditorList;
          SP_ScrollInView(SP_Store_Line(TokensStr));
          Error.Code := SP_ERR_NO_ERROR;
          SP_DisableDisplay;
          SP_ResizeCommandWindow(UsrCWW, UsrCWH, False);
          SP_MoveCommandWindow(UsrCWX, UsrCWY, False);
          SP_EnableDisplay;
          SP_PlaySystem(OKCHAN, OKSNDBANK);
        End Else Begin
          SP_HideEditorWindows;
          SP_AddToHistory(EDITLINE);
          EDITLINE := '';
          Error.Line := -1;
          Error.Statement := 1;
          Error.Position := SP_FindStatement(@TokensStr, 1);
          COMMAND_TOKENS := TokensStr;
          NXTSTATEMENT := -1;
          NXTLINE := -1;
          SP_StackPtr := SP_StackStart;
          Tokens := @TokensStr;
          If PROGCHANGED Then Begin
            SP_DeleteIncludes;
            SP_PreParse(False, Error);
            PreParseErrorCode := Error.Code;
            PreParseErrorLine := Error.Line;
            PreParseErrorStatement := Error.Statement;
            Error.Code := SP_ERR_OK;
            PROGCHANGED := False;
          End Else
            PreParseErrorCode := SP_ERR_OK;
          PROGSTATE := SP_PR_RUN;
          BREAKSIGNAL := False;
          SetExceptionMask([exInvalidOp, exDenormalized, exZeroDivide, exOverflow, exUnderflow, exPrecision]);
          ClearFlags;
          OUTSET := EditorOutSet;
          SP_Interpreter(Tokens, Error.Position, Error, PreParseErrorCode);
          If (Error.Code <> SP_ERR_OK) And (Error.Code = PreParseErrorCode) Then Begin
            Error.Line := PreParseErrorLine;
            Error.Statement := PreParseErrorStatement;
            SP_CreateEditorWindows;
            SP_MakeEditorList;
          End;
        End;
      End Else Begin
        // Ok, not a proper line - is it an expression?
        // Expressions will return error 7 every time - invalid keyword. Of course, it could actually be
        // an invalid keyword...
        If Error.Code = SP_ERR_INVALID_KEYWORD Then Begin

          Error.Code := SP_ERR_OK;
          Error.Position := 1;
          TokensStr := SP_TokeniseLine(EDITLINE, True, False) + SP_TERMINAL_SEQUENCE;
          EDITLINE := SP_DeTokenise(TokensStr, Error.Position, False, False);
          Idx := Length(TokensStr) +1;
          Expr := SP_Convert_Expr(TokensStr, Error.Position, Error, -1) + SP_TERMINAL_SEQUENCE;
          TokensStr := TokensStr + Expr;
          If TokensStr[Error.Position] <> SP_TERMINAL_CHAR Then Error.Code := SP_ERR_SYNTAX_ERROR;

          If Error.Code = SP_ERR_OK Then Begin
            Error.Position := Idx;
            COMMAND_TOKENS := TokensStr;
            NXTSTATEMENT := -1;
            NXTLINE := -1;
            SP_StackPtr := SP_StackStart;
            Tokens := @TokensStr;
            If PROGCHANGED Then Begin
              SP_DeleteIncludes;
              SP_PreParse(False, Error);
              PreParseErrorCode := Error.Code;
              Error.Code := SP_ERR_OK;
              PROGCHANGED := False;
            End Else
              PreParseErrorCode := SP_ERR_OK;
            PROGSTATE := SP_PR_RUN;
            BREAKSIGNAL := False;
            SetExceptionMask([exInvalidOp, exDenormalized, exZeroDivide, exOverflow, exUnderflow, exPrecision]);
            ClearFlags;
            OUTSET := EditorOutSet;
            SP_SetDrawingWindow(SCREENBANK, EditorDefaultWindow);
            SP_Interpreter(Tokens, Error.Position, Error, PreParseErrorCode);
            SP_SetDrawingWindow(SCREENBANK, CommandWindow);
            If (Error.Code = SP_ERR_OK) And (LongWord(SP_StackPtr) = LongWord(SP_StackStart) + SizeOf(SP_StackItem)) Then Begin
              Error.Code := SP_ERR_NO_ERROR;
              SP_AddToHistory(EDITLINE);
              If SP_StackPtr^.OpType = SP_VALUE Then
                Expr := FloatToStr(SP_StackPtr^.Val) + ' '
              Else Begin
                Expr := SP_StackPtr^.Str;
                If Length(Expr) > 256 Then
                  Expr := Copy(Expr, 1, 256);
                Expr := '"' + Expr + '" ';
              End;
              EDITLINE := '';
              SP_PlaySystem(OKCHAN, OKSNDBANK);
              SP_AddToHistory(Expr);
              For Idx := 1 To Length(Expr) Do Begin
                EDITLINE := Copy(Expr, 1, Idx);
                CURSORPOS := Length(EDITLINE);
                CURSORCHAR := 32;
                SP_EditorDisplayEditLine;
                SP_StackPtr := SP_StackStart;
                Inc(SP_StackPtr);
                SP_StackPtr^.Val := -1;
                SP_Interpret_WAIT(pInfo);
              End;
              EDITERROR := False;
              CURSORFG := 15;
              CURSORBG := 4;
              LocalFlashState := FLASHSTATE;
              While (LASTKEY <> 0) Or (MOUSEBTN <> 0) Do CB_Yield;
              LASTKEY := 0;
              While (LASTKEY = 0) AND (MOUSEBTN = 0) Do Begin
                If LocalFlashState <> FLASHSTATE Then Begin
                  If SCREENID <> COMMANDWINDOW Then
                    SP_SetDrawingWindow(SCREENBANK, COMMANDWINDOW);
                  SP_Display_Cursor;
                  LocalFlashState := FLASHSTATE;
                End;
                CB_Yield;
              End;
              MOUSEBTN := 0;
              EDITLINE := '';
              CURSORPOS := Length(EDITLINE) +1;
              CURSORCHAR := 32;
              SP_DisableDisplay;
              SP_ResizeCommandWindow(UsrCWW, UsrCWH, False);
              SP_MoveCommandWindow(UsrCWX, UsrCWY, False);
              SP_EnableDisplay;
              SP_PlaySystem(CLICKCHAN, CLICKBANK);
            End Else Begin
              If (Error.Code = SP_ERR_OK) And (SP_StackPtr = SP_StackStart) Then Begin
                Error.Code := SP_ERR_SYNTAX_ERROR;
                EDITERRORPOS := Min(Error.Position, Length(EDITLINE));
                EDITERROR := True;
                SP_PlaySystem(ERRORCHAN, ERRSNDBANK);
              End;
              Error.Line := -1;
              Error.Statement := 1;
              SP_HideEditorWindows;
              Goto RuntimeExit;
            End;
          End Else
            EDITLINE := SP_DeTokenise(TokensStr, Error.Position, False, False);
        End;
        If (Error.Code <> SP_ERR_OK) And (Error.Code <> SP_ERR_NO_ERROR) Then Begin
          EDITERRORPOS := Error.Position;
          EDITERROR := True;
          SP_PlaySystem(ERRORCHAN, ERRSNDBANK);
        End;
      End;

    End Else Begin
      SP_CycleEditorWindows;
      Error.Code := SP_ERR_NO_ERROR;
    End;

    RunTimeExit:

  Until QUITMSG;

End;

Procedure SP_GetUserInput;
Var
  Finished, Changed: Boolean;
  KeyChar: Byte;
  RepeatLen: LongWord;
  LocalFlashState: Integer;
Begin

  SP_DrawBatteryStatus;
  SP_EnableDisplay;

  CURSORFG := 15;
  If EDITERROR Then Begin
    CURSORBG := 10;
    CURSORPOS := EDITERRORPOS;
  End Else Begin
    CURSORBG := 9;
    If EDITLINE = '' Then CURSORPOS := 1;
  End;

  SYSTEMSTATE := SS_EDITOR;
  Finished := False;
  Changed := True;
  KeyChar := 0;
  RepeatLen := REPDEL;

  While Not (Finished or QUITMSG) Do Begin

    If Changed Then Begin
      SP_EditorDisplayEditLine;
      Changed := False;
    End;

    // Wait for a key - alphanumeric, of course. Shift doesn't count.
    // Also handle mouse events

    LocalFlashState := FLASHSTATE;
    Repeat
      If LocalFlashState <> FLASHSTATE Then Begin
        If SCREENID <> COMMANDWINDOW Then
          SP_SetDrawingWindow(SCREENBANK, COMMANDWINDOW);
        SP_Display_Cursor;
        LocalFlashState := FLASHSTATE;
      End;
      CB_Yield;
      If LASTKEY = 0 Then KeyChar := 0;
      If QUITMSG Then Exit;
    Until M_DOWNFLAG or M_UPFLAG or M_MOVEFLAG or M_WHEELUPFLAG or M_WHEELDNFLAG or (LASTKEY <> 0);

    If LastKey <> 0 Then Begin

      DEBUGVAL := LASTKEY;

      If LASTKEY = K_RETURN Then Begin
        SYSTEMSTATE := SS_IDLE;
        Exit;
      End;

      // Is this key the same as the last one?

      If KeyChar = LASTKEY Then Begin

        // Yes - make it repeat if necessary.

        If FRAMES - REPCOUNT >= RepeatLen Then Begin
          RepeatLen := REPPER;
          REPCOUNT := FRAMES;
          SP_EditorPerformEdit(LASTKEY);
          Changed := True;
        End;

      End Else Begin

        // No - This is a new key

        SP_EditorPerformEdit(LASTKEY);
        RepeatLen := REPDEL;
        REPCOUNT := FRAMES;
        KeyChar := LASTKEY;
        Changed := True;

      End;

    End Else Begin

      If M_DOWNFLAG Then SP_EditorHandleMouseDown(MOUSEX, MOUSEY);
      If M_UPFLAG Then SP_EditorHandleMouseUp(MOUSEX, MOUSEY);
      If M_MOVEFLAG Then SP_EditorHandleMouseMove(MOUSEX, MOUSEY);
      If M_WHEELUPFLAG Then SP_EditorHandleMouseWheel(True, MOUSEX, MOUSEY);
      If M_WHEELDNFLAG Then SP_EditorHandleMouseWheel(False, MOUSEX, MOUSEY);

    End;

  End;

End;

Procedure SP_EditorHandleMouseDown(X, Y: Integer);
Var
  Font: Integer;
  Err: TSP_ErrorCode;
Begin

  // Handle mouse-down event

  // First swap in the editor font, so metrics are correct.

  Font := SP_SetEditorFont;

  If MouseBtn = 1 Then Begin
    If ((X >= COMMANDWINX) And (Y >= COMMANDWINY) And (X < COMMANDWINX + COMMANDWINW) And (Y < COMMANDWINY + COMMANDWINH)) or (EditorMouseControl in [11, 12]) Then Begin
      Dec(X, COMMANDWINX);
      Dec(Y, COMMANDWINY);
      If EditorMouseControl = 0 Then Begin
        If (X >= COMMANDSIZETLX) and (X < COMMANDSIZEBRX) And (Y >= COMMANDSIZETLY) And (Y < COMMANDSIZEBRY) then Begin
          EditorMouseControl := 11;
          EditorMouseX := X;
          EditorMouseY := Y;
        End;
      End;
      If EditorMouseControl = 0 Then Begin
        If (Y >= 0) And (Y < EditorFontH +2) Then Begin
          EditorMouseControl := 12;
          EditorMouseX := MOUSEX;
          EditorMouseY := MOUSEY;
        End;
      End;
    End Else
      If SP_ListWindowShowing Then Begin
        If ((X >= LISTWINX) And (Y >= LISTWINY) And (X < LISTWINX + LISTWINW) And (Y < LISTWINY + LISTWINH)) or (EditorMouseControl in [1..7]) Then Begin
          Dec(X, LISTWINX);
          Dec(Y, LISTWINY);
          If (X >= 4) And (X < LISTVSCUPTLX) And (Y >= LISTVSCUPTLY) And (Y < LISTVSCDNBRY) Then Begin
            If EditorMouseControl = 0 Then Begin
              If (FRAMES - EditorClickTime < REPDEL) And (Abs(X - Integer(EditorClickX)) < 4) And (Abs(Y - Integer(EditorClickY)) < 4) Then Begin
                SP_EditorEditLine;
                M_DOWNFLAG := False;
              End Else Begin
                EditorClickX := X;
                EditorClickY := Y;
                M_DOWNFLAG := False;
                Y := Round(LISTTOPINDEX + ((Y-LISTVSCUPTLY-(EditorFontH/2))/EditorFontH));
                If Y < EditorList.Count Then Begin
                  EditorClickTime := FRAMES;
                  PROGLINE := LongWord(EditorList.Objects[Y]);
                  SP_AutoList;
                End;
                EditorMouseControl := 13;
              End;
            End;
          End;
          If EditorMouseControl = 0 Then Begin
            If (X >= LISTSIZETLX) and (X < LISTSIZEBRX) And (Y >= LISTSIZETLY) And (Y < LISTSIZEBRY) then Begin
              EditorMouseControl := 5;
              EditorMouseX := X;
              EditorMouseY := Y;
            End;
          End;
          If EditorMouseControl = 0 Then Begin
            If (Y >= 0) And (Y < EditorFontH +2) Then Begin
              EditorMouseControl := 6;
              EditorMouseX := MOUSEX;
              EditorMouseY := MOUSEY;
            End;
          End;
          If EditorMouseControl in [0, 1] Then Begin
            If LISTVSCUPENABLED Then // Check if user clicked down onto the listing window's "scroll up" arrow
              If (X >= LISTVSCUPTLX) And (X < LISTVSCUPBRX) And (Y >= LISTVSCUPTLY) and (Y < LISTVSCUPBRY) or (EditorMouseControl = 1) Then Begin
                If LISTTOPINDEX > 0 Then Begin
                  If Not EditorMouseDown or (EditorMouseDown And (EditorMouseDelay <= FRAMES)) Then Begin
                    If EditorMouseDown Then
                      EditorMouseDelay := FRAMES + REPPER
                    Else
                      EditorMouseDelay := FRAMES + REPDEL;
                    LISTTOPINDEX := Max(LISTTOPINDEX -1, 0);
                    SP_AutoList;
                    EditorMouseControl := 1;
                  End;
                End;
              End;
          End;
          If EditorMouseControl in [0, 2] Then Begin
            If LISTVSCDNENABLED Then // Check if user clicked down onto the listing window's "scroll down" arrow
              If (X >= LISTVSCDNTLX) And (X < LISTVSCDNBRX) And (Y >= LISTVSCDNTLY) and (Y < LISTVSCDNBRY) or (EditorMouseControl = 2) Then Begin
                If LISTTOPINDEX + LISTVISLINES < LISTLINECOUNT Then Begin
                  If Not EditorMouseDown or (EditorMouseDown And (EditorMouseDelay <= FRAMES)) Then Begin
                    If EditorMouseDown Then
                      EditorMouseDelay := FRAMES + REPPER
                    Else
                      EditorMouseDelay := FRAMES + REPDEL;
                    LISTTOPINDEX := LISTTOPINDEX +1;
                    SP_AutoList;
                    EditorMouseControl := 2;
                  End;
                End;
              End;
          End;
          If EditorMouseControl in [0, 3, 4] Then Begin
            If LISTVSTRENABLED Then Begin // Check if we've got a mousedown in the tracking region of the list window's vertical scrollbar
              If (X >= LISTVSTRTLX) And (X < LISTVSTRBRX) And (Y >= LISTVSTRTLY) and (Y < LISTVSTRBRY) or (EditorMouseControl in [3, 4]) Then Begin
                // In the thumb?
                If ((Y >= LISTVTHP) And (Y < LISTVTHP + LISTVTHS)) and (EditorMouseControl in [0, 4]) Then Begin
                  M_DOWNFLAG := False;
                  EditorMouseControl := 4;
                  If Not EditorMouseDown Then Begin
                    EditorMouseY := Y;
                  End;
                End Else Begin
                  If EditorMouseControl in [0, 3] Then Begin
                    If Not EditorMouseDown or (EditorMouseDown And (EditorMouseDelay <= FRAMES)) Then Begin
                      If EditorMouseDown Then
                        EditorMouseDelay := FRAMES + REPPER
                      Else
                        EditorMouseDelay := FRAMES + REPDEL;
                      If Y < LISTVTHP Then
                        LISTTOPINDEX := LISTTOPINDEX - (LISTVISLINES -1)
                      Else If Y >= LISTVTHP + LISTVTHS Then
                        LISTTOPINDEX := LISTTOPINDEX + LISTVISLINES -1;
                      LISTTOPINDEX := Min(Max(LISTTOPINDEX, 0), LISTLINECOUNT - LISTVISLINES);
                      SP_AutoList;
                      EditorMouseControl := 3;
                    End;
                  End;
                End;
              End;
            End;
          End;
          If EditorMouseControl in [0, 7] Then Begin
            If LISTHSCUPENABLED Then // Check if user clicked down onto the listing window's "scroll left" arrow
              If (X >= LISTHSCUPTLX) And (X < LISTHSCUPBRX) And (Y >= LISTHSCUPTLY) and (Y < LISTHSCUPBRY) or (EditorMouseControl = 7) Then Begin
                If LISTLEFTINDEX > 0 Then Begin
                  If Not EditorMouseDown or (EditorMouseDown And (EditorMouseDelay <= FRAMES)) Then Begin
                    If EditorMouseDown Then
                      EditorMouseDelay := FRAMES + REPPER
                    Else
                      EditorMouseDelay := FRAMES + REPDEL;
                    LISTLEFTINDEX := LISTLEFTINDEX -1;
                    SP_AutoList;
                    EditorMouseControl := 7;
                  End;
                End;
              End;
          End;
          If EditorMouseControl in [0, 8] Then Begin
            If LISTHSCDNENABLED Then Begin // Check if user clicked down onto the listing window's "scroll right" arrow
              If (X >= LISTHSCDNTLX) And (X < LISTHSCDNBRX) And (Y >= LISTHSCDNTLY) and (Y < LISTHSCDNBRY) or (EditorMouseControl = 8) Then Begin
                If LISTLEFTINDEX + LISTVISCOLS < LISTCOLCOUNT Then Begin
                  If Not EditorMouseDown or (EditorMouseDown And (EditorMouseDelay <= FRAMES)) Then Begin
                    If EditorMouseDown Then
                      EditorMouseDelay := FRAMES + REPPER
                    Else
                      EditorMouseDelay := FRAMES + REPDEL;
                    LISTLEFTINDEX := LISTLEFTINDEX +1;
                    SP_AutoList;
                    EditorMouseControl := 8;
                  End;
                End;
              End;
            End;
          End;
          If EditorMouseControl in [0, 9, 10] Then Begin
            If LISTHSTRENABLED Then Begin // Check if we've got a mousedown in the tracking region of the list window's vertical scrollbar
              If (X >= LISTHSTRTLX) And (X < LISTHSTRBRX) And (Y >= LISTHSTRTLY) and (Y < LISTHSTRBRY) or EditorMouseDown Then Begin
                // In the thumb?
                If ((X >= LISTHTHP) And (X < LISTHTHP + LISTHTHS)) and (EditorMouseControl in [0, 10]) Then Begin
                  M_DOWNFLAG := False;
                  EditorMouseControl := 10;
                  If Not EditorMouseDown Then Begin
                    EditorMouseX := X;
                  End;
                End Else Begin
                  If EditorMouseControl in [0, 9] Then Begin
                    If Not EditorMouseDown or (EditorMouseDown And (EditorMouseDelay <= FRAMES)) Then Begin
                      If EditorMouseDown Then
                        EditorMouseDelay := FRAMES + REPPER
                      Else
                        EditorMouseDelay := FRAMES + REPDEL;
                      If X < LISTHTHP Then
                        LISTLEFTINDEX := LISTLEFTINDEX - (LISTVISCOLS -1)
                      Else
                        If X >= LISTHTHP + LISTHTHS Then
                          LISTLEFTINDEX := LISTLEFTINDEX + LISTVISCOLS -1;
                      LISTLEFTINDEX := Min(Max(LISTLEFTINDEX, 0), LISTCOLCOUNT - LISTVISCOLS);
                      SP_AutoList;
                      EditorMouseControl := 9;
                    End;
                  End;
                End;
              End;
            End;
          End;
        End;
      End;

  End Else

    If MOUSEBTN = 4 Then Begin

    End;

  SP_SetSystemFont(Font, Err);
  EditorMouseDown := True;
  M_DOWNFLAG := False;
  M_UPFLAG := False;

End;

Procedure SP_EditorHandleMouseUp(X, Y: Integer);
Begin

  EditorMouseDown := False;
  EditorMouseControl := 0;
  M_DOWNFLAG := False;
  M_UPFLAG := False;

End;

Procedure SP_EditorHandleMouseMove(X, Y: Integer);
Var
  Dy, Dx, OldIndex, Font: Integer;
  Err: TSP_ErrorCode;
Begin

  Font := SP_SetEditorFont;

  If EditorMouseDown Then Begin
    If EditorMouseControl = 4 Then Begin // Vertical Track
      Dec(Y, LISTWINY);
      Dy := Y - EditorMouseY;
      OldIndex := Round(LISTTOPINDEX);
      LISTTOPINDEX := ((LISTVTHP - LISTVSTRTLY) + Dy) * ((LISTLINECOUNT - LISTVISLINES)/((LISTVSTRBRY - LISTVSTRTLY)-LISTVTHS));
      If OldIndex <> Round(LISTTOPINDEX) Then Begin
        LISTTOPINDEX := Min(Max(LISTTOPINDEX, 0), LISTLINECOUNT - LISTVISLINES);
        EditorMouseY := Y;
        SP_AutoList;
      End;
    End Else
      If EditorMouseControl = 5 Then Begin // Resize LIST window
        Dec(X, LISTWINX);
        Dec(Y, LISTWINY);
        Dx := X - EditorMouseX;
        Dy := Y - EditorMouseY;
        If (Dx <> 0) or (Dy <> 0) Then Begin
          SP_SetDirtyRect(LISTWINX, LISTWINY, LISTWINX + LISTWINW, LISTWINY + LISTWINH);
          Inc(LISTWINW, Dx);
          Inc(LISTWINH, Dy);
          LISTWINH := Max(LISTWINH, (EditorFontH * 6) +2);
          LISTWINW := Max(LISTWINW, (EditorFontW * 6) +2);
          EditorMouseX := X;
          EditorMouseY := Y;
          If LISTTOPINDEX + LISTVISLINES > LISTLINECOUNT Then
            LISTTOPINDEX := Max(LISTLINECOUNT - LISTVISLINES, 0);
          If LISTLEFTINDEX + LISTVISCOLS > LISTCOLCOUNT Then
            LISTLEFTINDEX := Max(LISTCOLCOUNT - LISTVISCOLS, 0);
          SP_DisableDisplay;
          SP_ResizeWindow(LISTWINDOW, LISTWINW, LISTWINH, -1, Err);
          SP_Decorate_Window(LISTWINDOW, 'Program listing - ' + PROGNAME, True);
          SP_MakeEditorList;
          SP_AutoList;
          SP_SetDirtyRect(LISTWINX, LISTWINY, LISTWINX + LISTWINW, LISTWINY + LISTWINH);
          SP_NeedDisplayUpdate := True;
          SP_EnableDisplay;
        End;
      End Else
        If EditorMouseControl = 6 Then Begin // Move LIST window
          Dx := X - EditorMouseX;
          Dy := Y - EditorMouseY;
          If (Dx <> 0) or (Dy <> 0) Then Begin
            SP_SetDirtyRect(LISTWINX, LISTWINY, LISTWINX + LISTWINW, LISTWINY + LISTWINH);
            Inc(LISTWINX, Dx);
            Inc(LISTWINY, Dy);
            EditorMouseX := X;
            EditorMouseY := Y;
            SP_MoveWindow(LISTWINDOW, LISTWINX, LISTWINY, Err);
            SP_SetDirtyRect(LISTWINX, LISTWINY, LISTWINX + LISTWINW, LISTWINY + LISTWINH);
            SP_NeedDisplayUpdate := True;
          End;
        End Else
          If EditorMouseControl = 10 Then Begin // Horizontal Track
            Dec(X, LISTWINX);
            Dx := X - EditorMouseX;
            OldIndex := Round(LISTLEFTINDEX);
            LISTLEFTINDEX := ((LISTHTHP - LISTHSTRTLX) + Dx) * ((LISTCOLCOUNT - LISTVISCOLS)/((LISTHSTRBRX - LISTHSTRTLX)-LISTHTHS));
            If OldIndex <> Round(LISTLEFTINDEX) Then Begin
              LISTLEFTINDEX := Min(Max(LISTLEFTINDEX, 0), LISTCOLCOUNT - LISTVISCOLS);
              EditorMouseX := X;
              SP_AutoList;
            End;
          End Else
            If EditorMouseControl = 11 Then Begin // Resize COMMAND window
              Dec(X, COMMANDWINX);
              Dec(Y, COMMANDWINY);
              Dx := X - EditorMouseX;
              Dy := Y - EditorMouseY;
              If (Dx <> 0) or (Dy <> 0) Then Begin
                SP_SetDirtyRect(COMMANDWINX, COMMANDWINY, COMMANDWINX + COMMANDWINW, COMMANDWINY + COMMANDWINH);
                Inc(COMMANDWINW, Dx);
                Inc(COMMANDWINH, Dy);
                COMMANDWINH := Min(DISPLAYHEIGHT - 8, Max(COMMANDWINH, (EditorFontH * 3) + 12));
                COMMANDWINW := Max(COMMANDWINW, (EditorFontW * 3) + 12);
                EditorMouseX := X;
                EditorMouseY := Y;
                SP_DisableDisplay;
                SP_ResizeCommandWindow(COMMANDWINW, COMMANDWINH, True);
                SP_Decorate_Window(COMMANDWINDOW, 'Direct command', True);
                SP_AutoList;
                SP_SetDirtyRect(COMMANDWINX, COMMANDWINY, COMMANDWINX + COMMANDWINW, COMMANDWINY + COMMANDWINH);
                SP_NeedDisplayUpdate := True;
                SP_EnableDisplay;
                UsrCWX := COMMANDWINX;
                UsrCWY := COMMANDWINY;
                UsrCWW := COMMANDWINW;
                UsrCWH := COMMANDWINH;
              End;
            End Else
              If EditorMouseControl = 12 Then Begin // Move COMMAND window
                Dx := X - EditorMouseX;
                Dy := Y - EditorMouseY;
                If (Dx <> 0) or (Dy <> 0) Then Begin
                  SP_SetDirtyRect(COMMANDWINX, COMMANDWINY, COMMANDWINX + COMMANDWINW, COMMANDWINY + COMMANDWINH);
                  Inc(COMMANDWINX, Dx);
                  Inc(COMMANDWINY, Dy);
                  EditorMouseX := X;
                  EditorMouseY := Y;
                  SP_MoveCommandWindow(COMMANDWINX, COMMANDWINY, True);
                  SP_SetDirtyRect(COMMANDWINX, COMMANDWINY, COMMANDWINX + COMMANDWINW, COMMANDWINY + COMMANDWINH);
                  SP_NeedDisplayUpdate := True;
                  UsrCWX := COMMANDWINX;
                  UsrCWY := COMMANDWINY;
                  UsrCWW := COMMANDWINW;
                  UsrCWH := COMMANDWINH;
                End;
              End;
  End;

  SP_SetSystemFont(Font, Err);
  M_MOVEFLAG := False;

End;

Procedure SP_EditorHandleMouseWheel(WheelUp: Boolean; X, Y: Integer);
Var
  LastTop: aFloat;
Begin

  LastTop := LISTTOPINDEX;
  If SP_ListWindowShowing Then
    If (X >= LISTWINX) And (Y >= LISTWINY) And (X < LISTWINX + LISTWINW) And (Y < LISTWINY + LISTWINH) Then Begin
      Dec(X, LISTWINX);
      Dec(Y, LISTWINY);
      If (X >= 4) And (X < LISTVSCUPTLX - 4) And (Y >= LISTVSCUPTLY) And (Y < LISTVSCDNBRY) Then Begin
        M_WHEELUPFLAG := False;
        M_WHEELDNFLAG := False;
        If WheelUp then
          LISTTOPINDEX := Max(LISTTOPINDEX -3, 0)
        Else
          LISTTOPINDEX := Max(0, Min(LISTTOPINDEX +3, LISTLINECOUNT - LISTVISLINES));
        If LISTTOPINDEX <> LastTop Then
          SP_AutoList;
      End;
    End;

End;

Procedure SP_ResizeCommandWindow(W, H: Integer; Redraw: Boolean);
Var
  Font: Integer;
  Win: pSP_Window_Info;
  Err: TSP_ErrorCode;
Begin

  SP_ResizeWindow(COMMANDWINDOW, W, H, -1, Err);
  SP_GetWindowDetails(COMMANDWINDOW, Win, SP_BankList[0], Err);

  COMMANDWINW := Win^.Width;
  COMMANDWINH := Win^.Height;

  Font := SP_SetEditorFont;
  COMMANDSIZETLX := COMMANDWINW - 4 - EditorFontW;
  COMMANDSIZETLY := COMMANDWINH - EditorFontH - 4;
  COMMANDSIZEBRX := COMMANDSIZETLX + EditorFontW;
  COMMANDSIZEBRY := COMMANDSIZETLY + EditorFontH;
  SP_SetSystemFont(Font, Err);

  SP_Decorate_Window(COMMANDWINDOW, 'Direct command', True);

  If Redraw Then
    SP_EditorDisplayEditLine;

End;

Procedure SP_MoveCommandWindow(X, Y: Integer; Redraw: Boolean);
Var
  Win: pSP_Window_Info;
  Err: TSP_ErrorCode;
Begin

  SP_MoveWindow(COMMANDWINDOW, X, Y, Err);
  SP_GetWindowDetails(COMMANDWINDOW, Win, SP_BankList[0], Err);
  COMMANDWINX := Win^.Left;
  COMMANDWINY := Win^.Top;
  If Redraw Then
    SP_EditorDisplayEditLine;

End;

Procedure SP_EditorDisplayEditLine;
Var
  Idx, WorkW, WorkH, NewW, NewH, NewX, NewY, TLen: Integer;
  EditLen, CPos, X, Y, WindowID, Font: Integer;
  CText, EL_Text: aString;
  Err: TSP_ErrorCode;
Begin

  If Not CCOMMANDWINDOW Then Exit;

  WindowID := SCREENID;
  T_INK := 0;
  T_OVER := 0;

  Font := SP_SetEditorFont;
  SP_SetDrawingWindow(SCREENBANK, COMMANDWINDOW);

  SP_DisableDisplay;

  // Get the edit line length

  Idx := 1;
  EditLen := 0;
  While Idx <= Length(EditLine) Do Begin
    If EDITLINE[Idx] >= ' ' Then Inc(EditLen);
    Inc(Idx);
  End;
  EditLen := Max(EditLen, 1);
  If CURSORPOS >= EditLen Then Inc(EditLen);

  // Figure out how much screen real estate we have to play with - take off the title, the border (5 pixels l/r/b) and the sizegrip.

  NewX := COMMANDWINX;
  NewY := COMMANDWINY;
  NewW := COMMANDWINW;
  WorkW := COMMANDWINW - (EditorFontW + 2 + 4);
  WorkH := COMMANDWINH - ((EditorFontH * 2) + 4);

  If EditLen > ((WorkW Div EditorFontW) * (WorkH Div EditorFontH)) Then Begin

    // Extend the window downwards, and if necessary move it upwards.

    NewH := (Ceil(EditLen/(WorkW Div EditorFontW)) * EditorFontW) + ((EditorFontH * 2) + 4);

    If NewY + NewH > DISPLAYHEIGHT - 4 Then
      NewY := (DISPLAYHEIGHT - 4) - NewH;

    SP_MoveCommandWindow(NewX, NewY, False);
    SP_ResizeCommandWindow(NewW, NewH, False);
    WorkW := COMMANDWINW - (EditorFontW + 2 + 4);

  End;

  SP_FillRect(1, EditorFontH +6, COMMANDWINW -2, COMMANDWINH - (EditorFontH +7), 7);
  SP_TextOut(-1, COMMANDWINW - 4 - EditorFontW, COMMANDWINH - EditorFontH - 4, EdSc + #250, 0, 7, True);
  Y := EditorFontH + 6;

  EL_Text := EDITLINE;
  If EL_Text = '' Then Begin
    EL_Text := ' ';
    CURSORPOS := 1;
  End;

  CPos := CURSORPOS;
  If CPos = Length(EL_Text) +1 Then
    EL_Text := EL_Text + ' ';

  While EL_Text <> '' Do Begin
    TLen := WorkW Div EditorFontW;
    X := 4; CText := ''; Idx := 1;
    While (TLen > 0) and (Idx <= Length(EL_Text)) Do Begin
      CText := CText + EL_Text[Idx];
      If EL_Text[Idx] >= ' ' Then
        Dec(TLen);
      Inc(Idx);
    End;
    SP_PRINT(-1, X, Y, cPos + Length(EdSc), EdSc + CText, 0, 7, Err);
    Dec(CPos, Length(CText));
    EL_Text := Copy(EL_Text, Length(CText) +1, Length(EL_Text));
    Inc(Y, EditorFontH);
  End;

  SP_SetSystemFont(Font, Err);
  SP_SetDrawingWindow(SCREENBANK, WindowID);

  SP_EnableDisplay;
  SP_NeedDisplayUpdate := True;

End;

Function  SP_FindLineNum(LineNum: Integer): Integer;
Var
  Idx: Integer;
Begin

  Idx := 0;
  While Idx < EditorList.Count Do
    If Integer(EditorList.Objects[Idx]) >= LineNum Then
      Break
    Else
      Inc(Idx);

  If Idx < EditorList.Count Then
    Result := Idx
  Else
    Result := -1;

End;

Procedure SP_BringToEditor(LineNum, Statement: Integer; Var Error: TSP_ErrorCode);
Var
  Idx: Integer;
  uLabel: aString;
  tLabel: TSP_Label;
Begin

  If EditorList.Count = 0 Then SP_MakeEditorList;

  If (LineNum = 0) And (Statement = 0) Then Begin
    SP_SetWindowVisible(LISTWINDOW, True, Error);
    If EDITLINE <> '' Then Begin
      Idx := 1;
      If EDITLINE[Idx] = '@' Then Begin
        // Label search
        Inc(Idx);
        uLabel := '@';
        While (Idx <= Length(EDITLINE)) And (EDITLINE[Idx] in ['0'..'9', 'A'..'Z', 'a'..'z', '_']) Do Begin
          uLabel := uLabel + EDITLINE[Idx];
          Inc(Idx);
        End;
        uLabel := Lower(uLabel);
        SP_BuildLabelList;
        tLabel := SP_FindLabel(uLabel, Error);
        If Error.Code = SP_ERR_OK Then Begin
          LineNum := pLongWord(@SP_Program[tLabel.DLine][2])^;
          Statement := tLabel.DSt -1;
          If LineNum >= 0 Then Begin
            Linenum := SP_FindLineNum(LineNum);
            LineNum := LongWord(EditorList.Objects[LineNum]);
            SP_ScrollInView(LineNum);
            PROGLINE := LineNum;
          End;
        End Else Begin
          EDITERRORPOS := 1;
          EDITERROR := True;
          SP_PlaySystem(ERRORCHAN, ERRSNDBANK);
          Exit;
        End;
      End Else Begin
        // Line number (optional statement after ":") search
        While (Idx <= Length(EDITLINE)) And (EDITLINE[Idx] in ['0'..'9']) Do Begin
          LineNum := (LineNum * 10) + Ord(EDITLINE[Idx]) - 48;
          Inc(Idx);
        End;
        If EDITLINE[Idx] = ':' Then Begin
          Inc(Idx);
          While (Idx <= Length(EDITLINE)) and (EDITLINE[Idx] in ['0'..'9']) Do Begin
            Statement := (Statement * 10) + Ord(EDITLINE[Idx]) - 48;
            Inc(Idx);
          End;
        End;
        If Idx > Length(EDITLINE) Then Begin
          Linenum := SP_FindLineNum(LineNum);
          If LineNum >= 0 Then Begin
            LineNum := LongWord(EditorList.Objects[LineNum]);
            SP_ScrollInView(LineNum);
            PROGLINE := LineNum;
          End;
        End;
      End;
    End;
  End Else Begin
    Linenum := SP_FindLineNum(LineNum);
    If LineNum >= 0 Then Begin
      LineNum := LongWord(EditorList.Objects[LineNum]);
      SP_ScrollInView(LineNum);
      PROGLINE := LineNum;
    End;
  End;
  SP_EditorEditLine;
  If Statement > 1 Then Begin
    Idx := CURSORPOS;
    While (Idx <= Length(EDITLINE)) Do Begin
      If EDITLINE[Idx] = ':' Then Begin
        Dec(Statement);
        If Statement = 1 Then Break;
      End;
      Inc(Idx);
    End;
    If Idx < Length(EDITLINE) Then Begin
      Inc(Idx);
      While (Idx < Length(EDITLINE)) And (EDITLINE[Idx] <= ' ') Do
        Inc(Idx);
      CURSORPOS := Idx;
      If (Not CCOMMANDWINDOW) or (COMMANDWINDOW = -1) Then SP_CreateEditorWindows;
      SP_EditorDisplayEditLine;
    End;
  End;
End;

Procedure SP_EditorPerformEdit(Char: Byte);
Var
  NewChar: Byte;
  LineIdx, Idx, Cnt, LineNum, Statement: Integer;
  Error: TSP_ErrorCode;
Begin

  // Test for control keys, then convert to ASCII using the
  // KEYBOARDSTATE sysvar.

  NewChar := SP_DecodeKey(Char, False);

  {$IFDEF SPECCYKEYS}
  If (Char = K_1) And (KEYSTATE[K_SHIFT] = 1) Then Begin
    Char := K_TAB;
    NewChar := 0;
  End;
  {$ENDIF}

  If (NewChar = 0) And (LASTKEYCHAR <> #1) Then Begin

    Case Char of

      K_F1..K_F9:
        Begin
          If KEYSTATE[K_CONTROL] = 1 Then Begin
            If KEYSTATE[K_SHIFT] = 1 Then Begin
              If EditorMarks[Char - K_F1] = LongWord(PROGLINE) Then
                EditorMarks[Char - K_F1] := 0
              Else
                EditorMarks[Char - K_F1] := LongWord(PROGLINE);
            End Else Begin
              PROGLINE := EditorMarks[Char - K_F1];
              SP_ScrollInView(EditorMarks[Char - K_F1]);
            End;
            SP_AutoList;
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;


      K_TAB:
        Begin
          SP_BringToEditor(0, 0, Error);
          If Error.Code <> SP_ERR_OK Then Begin
            CURSORPOS := 1;
            CURSORFG := 15;
            CURSORBG := 10;
            Exit;
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_LEFT:
        Begin
          If CURSORPOS > 1 Then
            If KEYSTATE[K_CONTROL] = 1 Then Begin
              If EDITLINE[CURSORPOS -1] = ' ' Then Dec(CURSORPOS);
              While (CURSORPOS > 1) And (EDITLINE[CURSORPOS] = ' ') Do Dec(CURSORPOS);
              While (CURSORPOS > 1) and (EDITLINE[CURSORPOS] <> ' ') Do Dec(CURSORPOS);
              If CURSORPOS > 1 Then
                Inc(CURSORPOS);
            End Else Begin
              Dec(CURSORPOS);
            End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_RIGHT:
        Begin
          If CURSORPOS <= Length(EDITLINE) Then
            If KEYSTATE[K_CONTROL] = 1 Then Begin
              While (CURSORPOS < Length(EDITLINE)+1) and (EDITLINE[CURSORPOS] = ' ') Do Inc(CURSORPOS);
              While (CURSORPOS < Length(EDITLINE)+1) and (EDITLINE[CURSORPOS] <> ' ') Do Inc(CURSORPOS);
              If EDITLINE[CURSORPOS] = ' ' Then Inc(CURSORPOS);
            End Else Begin
              Inc(CURSORPOS);
            End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_UP:
        Begin
          If KEYSTATE[K_CONTROL] = 0 Then Begin
            If KEYSTATE[K_SHIFT] = 0 Then Begin
              Cnt := (COMMANDWINW - (10+EditorFontW)) Div EditorFontW;
              While (Cnt > -1) And (CURSORPOS > 1) Do Begin
                Dec(CURSORPOS);
                Dec(Cnt);
              End;
            End Else Begin
              If Length(EditorHistory) > 0 Then Begin
                If HistoryPos = Length(EditorHistory) Then Begin
                  SetLength(EditorHistory, Length(EditorHistory) +1);
                  EditorHistory[Length(EditorHistory) -1] := EDITLINE;
                End;
                If HistoryPos > 0 Then Begin
                  Dec(HistoryPos);
                  EDITLINE := EditorHistory[HistoryPos];
                  CURSORPOS := Length(EDITLINE)+1;
                  SP_EditorDisplayEditLine;
                End;
              End;
            End;
          End Else Begin
            LineIdx := SP_FindLine(PROGLINE, False);
            If LineIdx > -1 Then Begin
              If LineIdx > 0 Then Dec(LineIdx);
              PROGLINE := pLongWord(@SP_Program[LineIdx][2])^;
              SP_ScrollInView(PROGLINE);
              SP_AutoList;
            End Else Begin
              While (PROGLINE > 0) And (SP_FindLine(PROGLINE, False) = -1) Do
                Dec(PROGLINE);
              LineIdx := SP_FindLine(PROGLINE, False);
              If LineIdx >= 0 Then Begin
                PROGLINE := pLongWord(@SP_Program[LineIdx][2])^;
                SP_ScrollInView(PROGLINE);
                SP_AutoList;
              End;
            End;
            SP_SetWindowVisible(LISTWINDOW, True, Error);
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_DOWN:
        Begin
          If KEYSTATE[K_CONTROL] = 0 Then Begin
            If KEYSTATE[K_SHIFT] = 0 Then Begin
              Cnt := (COMMANDWINW - (10+EditorFontW)) Div EditorFontW;
              While (Cnt > -1) And (CURSORPOS < Length(EDITLINE)+1) Do Begin
                Inc(CURSORPOS);
                Dec(Cnt);
              End;
            End Else Begin
              If Length(EditorHistory) > 0 Then Begin
                If HistoryPos < Length(EditorHistory) -1 Then Begin
                  Inc(HistoryPos);
                  EDITLINE := EditorHistory[HistoryPos];
                  CURSORPOS := Length(EDITLINE)+1;
                  SP_EditorDisplayEditLine;
                  If HistoryPos = Length(EditorHistory) -1 Then
                    SetLength(EditorHistory, Length(EditorHistory) -1);
                End;
              End;
            End;
          End Else Begin
            LineIdx := SP_FindLine(PROGLINE, False);
            If LineIdx > -1 Then Begin
              If INCLUDEFROM > -1 Then Begin
                If LineIdx < INCLUDEFROM -1 Then
                  Inc(LineIdx)
              End Else
                If LineIdx < SP_Program_Count -1 Then
                  Inc(LineIdx);
              PROGLINE := pLongWord(@SP_Program[LineIdx][2])^;
              SP_ScrollInView(PROGLINE);
              SP_AutoList;
            End;
            SP_SetWindowVisible(LISTWINDOW, True, Error);
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_HOME:
        Begin
          If KEYSTATE[K_CONTROL] = 1 Then
            CURSORPOS := 1
          Else
            Dec(CURSORPOS, (CURSORX -2) Div EditorFontW);
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_END:
        Begin
          If KEYSTATE[K_CONTROL] = 1 Then
            CURSORPOS := Length(EDITLINE)+1
          Else Begin
            Cnt := CURSORX;
            Idx := CURSORPOS;
            While Cnt < COMMANDWINW -18 Do Begin
              If Idx <= Length(EDITLINE) Then Begin
                If EDITLINE[Idx] >= ' ' Then Begin
                  Inc(Idx);
                  Inc(Cnt, EditorFontW);
                End Else
                  Inc(Idx);
              End Else Begin
                Inc(Idx, 2);
                Break;
              End;
            End;
            CURSORPOS := Idx -2;
            If CURSORPOS > Length(EDITLINE) Then
              CURSORPOS := Length(EDITLINE)+1;
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_PRIOR:
        Begin
          If KEYSTATE[K_CONTROL] = 1 Then Begin
            LISTTOPINDEX := Max(0, LISTTOPINDEX - (LISTVISLINES -1));
            SP_AutoList;
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_NEXT:
        Begin
          If KEYSTATE[K_CONTROL] = 1 Then Begin
            LISTTOPINDEX := LISTTOPINDEX + (LISTVISLINES -1);
            SP_AutoList;
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_BACK:
        Begin
          If CURSORPOS > 1 Then Begin
            EDITLINE := Copy(EDITLINE, 1, CURSORPOS -2) + Copy(EDITLINE, CURSORPOS, Length(EDITLINE));
            Dec(CURSORPOS);
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_DELETE:
        Begin
          If EDITLINE <> '' Then Begin
            If CURSORPOS <= Length(EDITLINE) Then
              EDITLINE := Copy(EDITLINE, 1, CURSORPOS -1) + Copy(EDITLINE, CURSORPOS +1, Length(EDITLINE));
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_INSERT:
        Begin
          INSERT := Not INSERT;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      K_ESCAPE:
        Begin
          EDITLINE := '';
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

    End;

    If CURSORPOS > Length(EDITLINE) Then
      CURSORCHAR := 32
    Else
      CURSORCHAR := Ord(EDITLINE[CURSORPOS]);

  End Else Begin

    If (KEYSTATE[K_CONTROL] = 1) And (KEYSTATE[K_ALT] = 0) Then Begin
      If INSERT Then Begin
        EDITLINE := Copy(EDITLINE, 1, CURSORPOS -1) + aChar(Byte(NewChar)+128) + Copy(EDITLINE, CURSORPOS, Length(EDITLINE));
        Inc(CursorPos);
      End Else Begin
        EDITLINE := Copy(EDITLINE, 1, CURSORPOS -1) + aChar(Byte(NewChar)+128) + Copy(EDITLINE, CURSORPOS +1, LENGTH(EDITLINE));
        Inc(CURSORPOS);
      End;
      SP_PlaySystem(CLICKCHAN, CLICKBANK);
    End Else Begin
      If INSERT Then
        EDITLINE := Copy(EDITLINE, 1, CURSORPOS -1) + aChar(NewChar) + Copy(EDITLINE, CURSORPOS, Length(EDITLINE))
      Else Begin
        EDITLINE := Copy(EDITLINE, 1, CURSORPOS -1) + aChar(NewChar) + Copy(EDITLINE, CURSORPOS +1, LENGTH(EDITLINE));
      End;
      Inc(CURSORPOS);
      SP_PlaySystem(CLICKCHAN, CLICKBANK);
    End;

  End;

  // When a key is pressed, clear the error state and set the cursor back to
  // the correct colours.

  EDITERROR := False;
  CURSORFG := 15;
  CURSORBG := 9;

End;

Procedure SP_EditorEditLine;
Var
  LineIdx, LineNum, cPos: Integer;
  LineNumStr: aString;
Begin

  // Pulls the line number specified by the sysvar PROGLINE into the edit line,
  // and sets the edit cursor to just after the line number.

  LineIdx := SP_FindLine(PROGLINE, False);
  If LineIdx >= 0 Then Begin
    If HistoryPos = Length(EditorHistory) Then Begin
      SetLength(EditorHistory, Length(EditorHistory) +1);
      EditorHistory[Length(EditorHistory) -1] := EDITLINE;
    End;
    EDITLINE := SP_DeTokenise(SP_Program[LineIdx], cPos, False, False);
    LineNum := 0;
    cPos := 1;
    While EDITLINE[cPos] in ['0'..'9'] Do Begin
      LineNum := (LineNum * 10) + (Ord(EDITLINE[cPos]) - 48);
      Inc(cPos);
    End;
    While EDITLINE[cPos] <= ' ' Do Inc(cPos);
    LineNumStr := IntToString(LineNum);
    CURSORPOS := cPos;
    SP_ScrollInView(LineNum);
    SP_EditorDisplayEditLine;
  End;

End;

Function SP_ListWindowShowing: Boolean;
Var
  Error: TSP_ErrorCode;
Begin
  Result := SP_WindowVisible(LISTWINDOW, Error);
End;

Procedure SP_EditorError(Error: TSP_ErrorCode);
Var
  Err: TSP_ErrorCode;
  ErrorText, Text, Title: aString;
  ErrorFPS: aFloat;
  ERRORWINDOW, WinW, WinH, WinX, WinY, MaxW, Lines, Cnt, Idx, MaxLen, bInk, bOver,
  Font, Window, ErrorDRPOSX, ErrorDRPOSY, ErrorPRPOSX, ErrorPRPOSY, ErrorFRAME_MS: Integer;
Begin

  Window := SCREENID;
  ErrorFPS := FPS;
  ErrorFRAME_MS := FRAME_MS;

  Font := SP_SetEditorFont;

  // Turn off ON ERROR - we don't want this to trigger now, it should have done it before if at all.
  // The EVERY system is also turned off, as the system should halt completely here.

  ERROR_LineNum := -1;
  SP_ClearEvery;

  SCREENLOCK := False;
  SYSTEMSTATE := SS_ERROR;
  EditorOutSet := OUTSET;
  OUTSET := False;

  // Get the error Text

  If Error.Line >= -1 Then Begin
    If Error.Line = -1 Then Error.Line := 0 Else Error.Line := pLongWord(@SP_Program[Error.Line][2])^;
    If Error.Code = 51 Then Begin
      If SP_KeyWordID < 4000 Then
        Text := IntToString(Error.Code)+' '+ErrorMessages[Error.Code] + SP_KEYWORDS[SP_KeyWordID - SP_KeyWord_Base] + ', ' + IntToString(Error.Line)+':'+IntToString(Error.Statement)
      Else
        Text := IntToString(Error.Code)+' '+ErrorMessages[Error.Code] + IntToString(SP_KeyWordID) + ', ' + IntToString(Error.Line)+':'+IntToString(Error.Statement);
    End Else
      Text := IntToString(Error.Code)+' '+ErrorMessages[Error.Code] + ', ' + IntToString(Error.Line)+':'+IntToString(Error.Statement);
  End Else
    {$IFDEF PANDORA}
    Text := #16#15#0#0#0#229#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237+
            #237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#233#16#0#0#0#0#13' '+
            #17#7#0#0#0#16#15#0#0#0#230#16#0#0#0#0#17#15#0#0#0'PandaBAS v'+BUILDSTR+'                             '#17#7#0#0#0#16#15#0#0#0#235+#13+
            #17#7#0#0#0+' '#17#7#0#0#0#16#15#0#0#0#230#16#0#0#0#0#17#15#0#0#0#127+' 2010 - 2014 by Paul Dunn.                 '#17#7#0#0#0#16#15#0#0#0#235#13+
            #17#7#0#0#0+' '#16#15#0#0#0#226#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228+
            #228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#227#16#0#0#0#0#13#13' '+
            #17#7#0#0#0' PandaBAS is currently ALPHA quality.'#13'  The author will not be held liable for any'#13+
            '  damage or loss caused by the use of this'#13'  software'#13'  ';
    {$ELSE}
    Text := #16#15#0#0#0#229#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237+
            #237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#237#233#16#0#0#0#0#13' '+
            #17#7#0#0#0#16#15#0#0#0#230#16#0#0#0#0#17#15#0#0#0'SpecBAS v'+BUILDSTR+'                              '#17#7#0#0#0#16#15#0#0#0#235+#13+
            #17#7#0#0#0+' '#17#7#0#0#0#16#15#0#0#0#230#16#0#0#0#0#17#15#0#0#0#127+' 2010 - 2014 by Paul Dunn.                 '#17#7#0#0#0#16#15#0#0#0#235#13+
            #17#7#0#0#0+' '#16#15#0#0#0#226#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228+
            #228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#228#227#16#0#0#0#0#13#13' '+
            {$IFDEF RASPI}
            #17#7#0#0#0' Based on Linux build by Piez.'#13#13'  SpecBAS is currently ALPHA quality software.'#13'  The author will not be held liable for any'#13+
            {$ELSE}
            #17#7#0#0#0' Linux build by Piez.'#13#13'  SpecBAS is currently ALPHA quality software.'#13'  The author will not be held liable for any'#13+
            {$ENDIF}
            '  damage or loss caused by the use of this'#13'  software'#13'  ';
   {$ENDIF}
  // Create a window.

  SP_DisableDisplay;

  Idx := 1;
  Cnt := 0;
  ErrorText := '';
  MaxW := DISPLAYWIDTH - (2 + EditorFontW + EditorFontW);
  While Idx <= Length(Text) Do Begin
    ErrorText := ErrorText + Text[Idx];
    If Text[Idx] >= ' ' Then
      Inc(Cnt);
    If Cnt > MaxW Div EditorFontW Then Begin
      If Text[Idx] = ' ' Then
        ErrorText[Length(ErrorText)] := #13
      Else Begin
        While (Idx > 1) and (Text[Idx] > ' ') Do Begin
          Dec(Idx);
          ErrorText := Copy(ErrorText, 1, Length(ErrorText) -1);
        End;
        ErrorText := ErrorText + #13;
      End;
      Cnt := 0;
    End;
    If Text[Idx] = #13 Then
      Cnt := 0;
    Inc(Idx);
  End;

  Idx := 1;
  Cnt := 0;
  Lines := 1;
  MaxLen := 0;
  While Idx < Length(ErrorText) Do Begin
    If ErrorText[Idx] >= ' ' Then
      Inc(Cnt);
    If ErrorText[Idx] = #13 Then Begin
      Inc(Lines);
      If Cnt > MaxLen Then
        MaxLen := Cnt;
      Cnt := 0;
    End;
    Inc(Idx);
  End;
  If Cnt > MaxLen Then
    MaxLen := Cnt;

  WinW := (MaxLen * EditorFontW) + 2 + 8;
  WinH := EditorFontH + 2 + 2 + 8 + (EditorFontH * Lines);

  SHOWLIST := False;
  If Error.Line < -1 Then
    Title := ' '
  Else
    If (Error.Code <> SP_ERR_OK) And (Error.Code <> SP_ERR_BREAK) And (Error.Code <> SP_ERR_STOP) Then Begin
      Title := 'SpecBAS error';
      If Error.Line > 0 Then Begin
        SP_ScrollInView(Error.Line);
        PROGLINE := Error.Line;
        SHOWLIST := True;
      End;
      WinW := DISPLAYWIDTH - 8;
    End Else Begin
      Title := 'SpecBAS message';
      WinW := DISPLAYWIDTH - 8;
    End;

  WinX := (DISPLAYWIDTH - WinW) Div 2;
  If Error.Line < -1 Then
    WinY := (DISPLAYHEIGHT - WinH) Div 2
  Else
    WinY := (DISPLAYHEIGHT - WinH) - 4;

  ERRORWINDOW := SP_Add_Window(SCREENBANK, WinX, WinY, WinW, WinH, -1, 8, 0, Error);
  SP_SetDrawingWindow(SCREENBANK, ERRORWINDOW);
  For Idx := 0 To 255 Do pSP_Window_Info(WINDOWPOINTER)^.Palette[Idx] := DefaultPalette[Idx];

  SP_FillRect(0, 0, WinW, WinH, 7);
  SP_Decorate_Window(ERRORWINDOW, Title, False);

  COVER := 0;
  T_INK := 0;
  T_OVER := 0;
  SP_TextOut(-1, EditorFontW, EditorFontH + 7, EdSc + ErrorText, 0, 7, True);
  SP_DrawRectangle(0, 0, WinW -1, WinH -1);

  SP_EnableDisplay;
  SCMINX := 0; SCMINY := 0; SCMAXX := DISPLAYWIDTH; SCMAXY := DISPLAYHEIGHT;
  SP_NeedDisplayUpdate := True;
  CauseUpdate := True;

  // Wait for any key

  While (LASTKEY <> 0) Or (MOUSEBTN <> 0) Do
    CB_Yield;

  LASTKEY := 0;
  While (LASTKEY = 0) AND (MOUSEBTN = 0) And Not QUITMSG Do
    CB_Yield;

  If QUITMSG Then Exit;

  SYSTEMSTATE := SS_IDLE;
  REPCOUNT := FRAMES;

  SP_DisableDisplay;
  SP_DeleteWindow(ERRORWINDOW, Err);
  SP_Stop_Sound;

  FPS := ErrorFPS;
  FRAME_MS := ErrorFRAME_MS;

  SP_SetSystemFont(Font, Err);
  SP_SetDrawingWindow(SCREENBANK, Window);
  SP_EnableDisplay;

  If LASTKEY = K_TAB Then
    If Error.Line > 0 Then
      If Error.Code > 0 Then Begin
        LASTKEY := 0;
        SP_BringToEditor(Error.Line, Error.Statement, Error);
      End;

  SP_NeedDisplayUpdate := True;
  CauseUpdate := True;

  If Error.Line >= -1 Then
    SP_PlaySystem(CLICKCHAN, CLICKBANK);

End;

Procedure SP_AddToHistory(Text: aString);
Var
  Idx, Idx2: Integer;
Begin

  While (Length(EditorHistory) > 0) And (EditorHistory[Length(EditorHistory) -1] = '') Do
    SetLength(EditorHistory, Length(EditorHistory) -1);
  While Text[1] <= ' ' Do
    Text := Copy(Text, 2, Length(Text));
  While Text[Length(Text)] <= ' ' Do
    Text := Copy(Text, 1, Length(Text) -1);
  For Idx := 0 To Length(EditorHistory) -1 Do Begin
    If Text = EditorHistory[Idx] Then Begin
      For Idx2 := Idx To Length(EditorHistory) -2 Do
        EditorHistory[Idx2] := EditorHistory[Idx2 +1];
      SetLength(EditorHistory, Length(EditorHistory) -1);
      Break;
    End;
  End;
  If (Length(EditorHistory) > 0) and (EditorHistory[Length(EditorHistory) -1] = Text) Then Exit;
  SetLength(EditorHistory, Length(EditorHistory) +1);
  EditorHistory[Length(EditorHistory) -1] := Text;
  HistoryPos := Length(EditorHistory);

End;

// INPUT routines. Very similar on the face of it to the Editor routines, but with subtle and important differences! Honest!

Function SP_GetInput(Var Error: TSP_ErrorCode): Boolean;
Var
  Finished, Changed: Boolean;
  KeyChar: Byte;
  RepeatLen: LongWord;
  LocalFlashState, X, Y, Top, PosX, PosY, Scrolls, Fg, Bg, dCnt: Integer;
  EL_Text, TempStr: aString;
Begin

  Result := False;
  X := 0; Y := 0;
  dCnt := DISABLECOUNT;
  DISABLECOUNT := 0;

  TempStr := SP_GrabCurrentWindow;

  CURSORFG := INPUTCSR1;
  CURSORBG := INPUTCSR2;
  INPUTLINE := SP_INPUTFromFORMAT;
  CURSORPOS := 1;
  If INFORMAT <> '' Then
    While (CURSORPOS <= Length(INPUTLINE)) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Inc(CURSORPOS);

  SYSTEMSTATE := SS_INPUT;
  Finished := False;
  Changed := True;
  KeyChar := 0;
  RepeatLen := REPDEL;
  LASTKEY := 0;
  Top := PRPOSY;

  While Not Finished Do Begin

    If Changed Then Begin

      EL_Text := INPUTLINE;
      If EL_Text = '' Then Begin
        EL_Text := ' ';
        CURSORPOS := 1;
      End;

      If CURSORPOS = Length(EL_Text) +1 Then
        EL_Text := EL_Text + ' ';

      X := PRPOSX; Y := PRPOSY;
      SCROLLCNT := 0;
      SP_PutCurrentWindow(TempStr);
      If SCREENBPP = 8 Then
        Scrolls := SP_PRINT(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error)
      Else
        Scrolls := SP_PRINT32(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error);
      While Scrolls > 0 Do Begin
        Dec(Y, FONTHEIGHT * Scrolls);
        SP_PutCurrentWindow(TempStr);
        If SCREENBPP = 8 Then
          Scrolls := SP_PRINT(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error)
        Else
          Scrolls := SP_PRINT32(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error);
      End;
      SP_SetDirtyRect(0, 0, DISPLAYWIDTH, DISPLAYHEIGHT);
      PRPOSX := X; PRPOSY := Y;
      Changed := False;

    End;

    LocalFlashState := FLASHSTATE;
    Repeat
      If LocalFlashState <> FLASHSTATE Then Begin
        If FlashState = 1 Then Begin
          Fg := CURSORFG; Bg := CURSORBG;
        End Else Begin
          Fg := CURSORBG; Bg := CURSORFG;
        End;

        PosX := PRPOSX; PosY := PRPOSY;
        If SCREENBPP = 8 Then Begin
          SP_TEXTOUT(-1, CURSORX, CURSORY, aChar(CURSORCHAR), Fg, Bg, True);
        End Else Begin
          SP_TEXTOUT32(-1, CURSORX, CURSORY, aChar(CURSORCHAR), Fg, Bg, True);
        End;
        PRPOSX := PosX;
        PRPOSY := PosY;
        LocalFlashState := FLASHSTATE;
      End;
      If QUITMSG Then Exit;
      CB_Yield;
      If LASTKEY = 0 Then KeyChar := 0;
    Until LASTKEY <> 0;

    If LASTKEY in [K_RETURN, K_ESCAPE] Then Begin
      SYSTEMSTATE := SS_IDLE;
      If LASTKEY = K_RETURN Then
        Result := True
      Else Begin
        INPUTLINE := '';
        Result := False;
      End;
      Break;
    End;

    // Is this key the same as the last one?

    If KeyChar = LASTKEY Then Begin

      // Yes - make it repeat if necessary.

      If FRAMES - REPCOUNT >= RepeatLen Then Begin
        RepeatLen := REPPER;
        REPCOUNT := FRAMES;
        SP_PerformINPUT(LASTKEY);
        Changed := True;
      End;

    End Else Begin

      // No - This is a new key

      SP_PerformINPUT(LASTKEY);
      RepeatLen := REPDEL;
      REPCOUNT := FRAMES;
      KeyChar := LASTKEY;
      Changed := True;

    End;

  End;

  SP_PutCurrentWindow(TempStr);
  If SCREENBPP = 8 Then
    Scrolls := SP_PRINT(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error)
  Else
    Scrolls := SP_PRINT32(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error);
  While Scrolls > 0 Do Begin
    Dec(Y, FONTHEIGHT * Scrolls);
    SP_PutCurrentWindow(TempStr);
    If SCREENBPP = 8 Then
      Scrolls := SP_PRINT(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error)
    Else
      Scrolls := SP_PRINT32(-1, X, Y, CURSORPOS, EL_Text, T_INK, T_PAPER, Error);
  End;
  SP_SetDirtyRect(0, 0, DISPLAYWIDTH, DISPLAYHEIGHT);
  PRPOSX := X; PRPOSY := Y;
  TempStr := '';

  DISABLECOUNT := dCnt;

  // Wait for the keyboard to go silent before ending the INPUT

  Repeat
    CB_Yield;
  Until LASTKEY = 0;

End;

Procedure SP_PerformINPUT(Char: Byte);
Var
  NewChar: Byte;
  LineIdx, Idx, Cnt, LineNum, Statement, PosM, PosT: Integer;
  Error: TSP_ErrorCode;
  nChar: aChar;
Begin

  NewChar := SP_DecodeKey(Char, False);

  If NewChar = 0 Then Begin

    Case Char of

      K_LEFT:
        Begin
          If CURSORPOS > 1 Then Begin
            If (INFORMAT = '') And (KEYSTATE[K_CONTROL] = 1) Then Begin
              If INPUTLINE[CURSORPOS -1] = ' ' Then Dec(CURSORPOS);
              While (CURSORPOS > 1) And (INPUTLINE[CURSORPOS] = ' ') Do Dec(CURSORPOS);
              While (CURSORPOS > 1) and (INPUTLINE[CURSORPOS] <> ' ') Do Dec(CURSORPOS);
              If CURSORPOS > 1 Then
                Inc(CURSORPOS);
            End Else Begin
              Dec(CURSORPOS);
              If INFORMAT <> '' Then Begin
                While (CURSORPOS > 1) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Dec(CURSORPOS);
                If CURSORPOS = 1 Then
                  While (CURSORPOS <= Length(INPUTLINE)) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Inc(CURSORPOS);
              End;
            End;
          End;
        End;

      K_RIGHT:
        Begin
          If CURSORPOS <= Length(INPUTLINE) Then Begin
            If (INFORMAT = '') And (KEYSTATE[K_CONTROL] = 1) Then Begin
              While (CURSORPOS < Length(INPUTLINE)+1) and (INPUTLINE[CURSORPOS] = ' ') Do Inc(CURSORPOS);
              While (CURSORPOS < Length(INPUTLINE)+1) and (INPUTLINE[CURSORPOS] <> ' ') Do Inc(CURSORPOS);
              If INPUTLINE[CURSORPOS] = ' ' Then Inc(CURSORPOS);
            End Else Begin
              Inc(CURSORPOS);
              If INFORMAT <> '' Then Begin
                While (CURSORPOS <= Length(INPUTLINE)) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Inc(CURSORPOS);
                If CURSORPOS = Length(INPUTLINE) Then
                  While (CURSORPOS > 1) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Dec(CURSORPOS);
              End;
            End;
          End;
        End;

      K_HOME:
        Begin
          CURSORPOS := 1;
          If INFORMAT <> '' Then
            While (CURSORPOS <= Length(INPUTLINE)) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Inc(CURSORPOS);
        End;

      K_END:
        Begin
          If INFORMAT <> '' Then Begin
            CURSORPOS := Length(INPUTLINE) +1;
            While (CURSORPOS > 1) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Dec(CURSORPOS);
          End Else
            CURSORPOS := Length(INPUTLINE)+1;
        End;

      K_BACK:
        Begin
          If CURSORPOS > 1 Then Begin
            If INFORMAT = '' Then Begin
              INPUTLINE := Copy(INPUTLINE, 1, CURSORPOS -2) + Copy(INPUTLINE, CURSORPOS, Length(INPUTLINE));
              Dec(CURSORPOS);
            End Else Begin
              Dec(CURSORPOS);
              While (CURSORPOS > 1) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Dec(CURSORPOS);
              If CURSORPOS = 1 Then
                While (CURSORPOS <= Length(INPUTLINE)) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Inc(CURSORPOS);
              INPUTLINE[CURSORPOS] := ' ';
            End;
          End;
        End;

      K_DELETE:
        Begin
          If INFORMAT = '' Then Begin
            If INPUTLINE <> '' Then Begin
              If CURSORPOS <= Length(INPUTLINE) Then
                INPUTLINE := Copy(INPUTLINE, 1, CURSORPOS -1) + Copy(INPUTLINE, CURSORPOS +1, Length(INPUTLINE));
            End;
          End Else Begin
            INPUTLINE[CURSORPOS] := ' ';
          End;
        End;

      K_INSERT:
        Begin
          If INFORMAT = '' Then
            INSERT := Not INSERT;
        End;

      K_ESCAPE:
        Begin
          INPUTLINE := '';
        End;

    End;

    If CURSORPOS > Length(INPUTLINE) Then
      CURSORCHAR := 32
    Else
      CURSORCHAR := Ord(INPUTLINE[CURSORPOS]);

  End Else Begin

    IF INFORMAT = '' Then Begin
      If KEYSTATE[K_CONTROL] = 1 Then Begin
        If INSERT Then Begin
          INPUTLINE := Copy(INPUTLINE, 1, CURSORPOS -1) + aChar(Byte(NewChar) + 128) + Copy(INPUTLINE, CURSORPOS, Length(INPUTLINE));
          Inc(CursorPos, 2);
        End Else Begin
          INPUTLINE := Copy(INPUTLINE, 1, CURSORPOS -1) + aChar(Byte(NewChar) + 128) + Copy(INPUTLINE, CURSORPOS +1, LENGTH(INPUTLINE));
          Inc(CURSORPOS, 2);
        End;
      End Else Begin
        If INSERT Then
          INPUTLINE := Copy(INPUTLINE, 1, CURSORPOS -1) + aChar(NewChar) + Copy(INPUTLINE, CURSORPOS, Length(INPUTLINE))
        Else Begin
          INPUTLINE := Copy(INPUTLINE, 1, CURSORPOS -1) + aChar(NewChar) + Copy(INPUTLINE, CURSORPOS +1, LENGTH(INPUTLINE));
        End;
        Inc(CURSORPOS);
      End;
    End Else If CURSORPOS <= Length(INPUTLINE) Then Begin
      nChar := aChar(NewChar);
      If INFORMAT[(CURSORPOS *2) -1] <> '\' Then
        Case INFORMAT[CURSORPOS *2] of
          'A':
            If nChar in ['a'..'z'] Then
              nChar := aChar(Ord(nChar) - 32)
            Else
              If Not (nChar in ['A'..'Z']) Then
                nChar := #0;
          'a':
            If nChar in ['A'..'Z'] Then
              nChar := aChar(Byte(nChar) + 32)
            Else
              If Not (nChar in ['a'..'z']) Then
                nChar := #0;
          '#':
            If Not (nChar in ['0'..'9']) Then
              nChar := #0;
          '<':
            If nChar in ['A'..'Z'] Then
              nChar := aChar(Ord(nChar) + 32);
          '>':
            If nChar in ['a'..'z'] Then
              nChar := aChar(Ord(nChar) - 32);
          '&':
            If Not (nChar in ['A'..'Z', 'a'..'z']) Then
              nChar := #0;
          '?':
            If Not (nChar in ['A'..'Z', 'a'..'z', '0'..'9']) Then
              nChar := #0;
        End;
      If nChar <> #0 Then Begin
        INPUTLINE := Copy(INPUTLINE, 1, CURSORPOS -1) + nChar + Copy(INPUTLINE, CURSORPOS +1, LENGTH(INPUTLINE));
        If CURSORPOS <= Length(INPUTLINE) Then Begin
          Inc(CURSORPOS);
          While (CURSORPOS <= Length(INPUTLINE)) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Inc(CURSORPOS);
          If CURSORPOS = Length(INPUTLINE) Then
            While (CURSORPOS > 1) And Not ((INFORMAT[CURSORPOS *2] in ['#', 'A', 'a', '<', '>', '&', '?', '*']) And (INFORMAT[(CURSORPOS *2)-1] <> '\')) Do Dec(CURSORPOS);
        End;
      End;
    End;

  End;

End;

Function SP_INPUTFromFORMAT: aString;
Var
  Idx: Integer;
Begin

  Result := '';

  Idx := 1;
  While Idx <= Length(INFORMAT) -2 Do Begin

    Case INFORMAT[Idx] Of
      '#', 'A', 'a', '<', '>', '&', '?', '*':
        Result := Result + ' ';
      '\':
        Begin
          Result := Result + INFORMAT[Idx +1];
        End;
    Else
      Result := Result + INFORMAT[Idx];
    End;

    Inc(Idx, 2);

  End;

End;

// Message box routines - info boxes, 128k-style menus etc

Function  SP_ShowMenu(Var Options: aString; cX, cY: Integer): Integer;
Var
  NumOptions, MaxWidth, WinX, WinY, WinW, WinH, Offset, Idx, StrLen,
  Window, WindowID, yOff, CurOption, RepeatLen, bWidth, bHeight, fW, fH: Integer;
  OptionList: TStringlist;
  cBlack, cWhite, cRed, cGreen, cCyan, cYellow, KeyChar: Byte;
  Err: TSP_ErrorCode;
  Win: pSP_Window_Info;
  Done: Boolean;
Begin

  KeyChar := 0;
  RepeatLen := 0;
  Result := -1;

  // Options in the string, in format length,string...
  // Centre the box on cX, cY in the current window.

  NumOptions := pLongWord(@Options[1])^;
  Optionlist := TStringlist.Create;

  MaxWidth := 0;
  Offset := SizeOf(LongWord) +1;
  For Idx := 1 To NumOptions Do Begin
    StrLen := pLongWord(@Options[Offset])^;
    OptionList.Add(Copy(String(Options), Offset + SizeOf(LongWord), StrLen));
    Inc(Offset, SizeOf(LongWord) + StrLen);
    If StrLen > MaxWidth Then
      MaxWidth := StrLen;
  End;

  // Now calculate the metrics of the window that we'll need
  // Width - one pixel border, FONTWIDTH-1 pixels space, text, FONTWIDTH-1 pixels space one pixel border
  // Height - one pixel border, optional title + one pixel separator, 4 pixels space, text, 4 pixels space, one pixel border.

  fW := Trunc(FONTWIDTH * EDFONTSCALEX);
  fH := Trunc(FONTHEIGHT * EDFONTSCALEY);

  bWidth := fW Div 2;
  bHeight := fH Div 2;
  If OptionList[0] <> '' Then Begin
    WinW := Max(((7 + Length(OptionList[0])) * fW) + bWidth, Min(DISPLAYWIDTH, (MaxWidth * fW) + (bWidth * 2) + 2));
    WinH := Min(DISPLAYHEIGHT, (NumOptions * fH) + (bHeight * 2) + 3);
  End Else Begin
    WinW := Min(DISPLAYWIDTH, (MaxWidth * fW) + (bWidth * 2) + 2);
    WinH := Min(DISPLAYHEIGHT, ((NumOptions -1) * fH) + (bHeight * 2) + 2);
  End;

  // Convert the cX,cY centre values to display space, as the menu uses a window, and thus the display coordinates.

  Inc(cX, SCREENX);
  Inc(cY, SCREENY);
  WinX := cX - (WinW Div 2);
  WinY := cY - (WinH Div 2);
  If WinX + WinW > DISPLAYWIDTH Then
    WinX := (DISPLAYWIDTH + WinW) Div 2;
  If WinY + WinH > DISPLAYHEIGHT Then
    WinY := (DISPLAYHEIGHT + WinH) Div 2;

  // Because we're not using the editor's technique of hijacking the palette, we need to choose
  // appropriate palette entries for this.

  cBlack := SP_Get_Nearest_Colour(0, 0, 0, -1);
  cRed := SP_Get_Nearest_Colour(255, 0, 0, -1);
  cGreen := SP_Get_Nearest_Colour(0, 255, 0, -1);
  cCyan := SP_Get_Nearest_Colour(0, 255, 255, -1);
  cYellow := SP_Get_Nearest_Colour(255, 255, 0, -1);
  cWhite := SP_Get_Nearest_Colour(255, 255, 255, -1);

  // Now create the window

  Window := SCREENID;
  WindowID := SP_Add_Window(SCREENBANK, WinX, WinY, WinW, WinH, -1, 8, 0, Err);
  SP_GetWindowDetails(WindowID, Win, SP_BankList[0], Err);
  SP_SetDrawingWindow(SCREENBANK, WindowID);
  For Idx := 0 To 255 Do pSP_Window_Info(WINDOWPOINTER)^.Palette[Idx] := DefaultPalette[Idx];

  // And draw the menu.

  T_INK := cBlack;
  SP_FillRect(0, 0, WinW, WinH, cWhite);
  SP_DrawRectangle(0, 0, WinW, WinH);

  If OptionList[0] <> '' Then Begin

    SP_FillRect(0, 0, WinW, fH + 2, cBlack);
    SP_TextOut(-1, 1, 1, EdSc + aString(OptionList[0]), 15, 0, True);
    SP_TextOut(-1, Win^.Width - ((fW * 6) + 2), 1, EdSc + aChar(16) + aChar(cRed) + #0#0#0 + aChar(17) + aChar(cBlack) + #0#0#0 + aChar(255) +
                                                          aChar(16) + aChar(cYellow) + #0#0#0 + aChar(17) + aChar(cRed) + #0#0#0 + aChar(255) +
                                                          aChar(16) + aChar(cGreen) + #0#0#0 + aChar(17) + aChar(cYellow) + #0#0#0 + aChar(255) +
                                                          aChar(16) + aChar(cCyan) + #0#0#0 + aChar(17) + aChar(cGreen) + #0#0#0 + aChar(255) +
                                                          aChar(16) + aChar(cBlack) + #0#0#0 + aChar(17) + aChar(cCyan) + #0#0#0 + aChar(255),
                                                          cWhite, 0, True);
    yOff := bHeight + 2;

  End Else

    yOff := (bHeight +1) - fH;

  // Now enter the main loop, drawing the options and waiting/responding on user input.
  // Allowed keys are up/down/enter/escape/home/end. Mouse may also be used, if the mouse is currently visible.
  // This is a modified version of the routines used by the INPUT systems above.

  Done := False;
  CurOption := 1;
  LASTKEY := 0;
  Repeat

    SP_DisableDisplay;
    For Idx := 1 to OptionList.Count -1 Do Begin
      If Idx = CurOption Then Begin
        SP_FillRect(bWidth +1, yOff + (Idx * fH), WinW - ((bWidth * 2) + 2), fH, cCyan);
        SP_TextOut(-1, bWidth +1, yOff + (Idx * fH), EdSc + aString(OptionList[Idx]), cBlack, cCyan, True)
      End Else Begin
        SP_FillRect(bWidth +1, yOff + (Idx * fH), WinW - ((bWidth * 2) + 2), fH, cWhite);
        SP_TextOut(-1, bWidth +1, yOff + (Idx * fH), EdSc + aString(OptionList[Idx]), cBlack, cWhite, True);
      End;
    End;
    SP_EnableDisplay;
    CB_Refresh_Display;

    Repeat
      CB_Yield;
      If LASTKEY = 0 Then KeyChar := 0;
    Until (LASTKEY <> 0) or QUITMSG;
    If QUITMSG Then Exit;

    If LASTKEY in [K_RETURN, K_ESCAPE] Then Begin
      If LASTKEY = K_RETURN Then
        Result := integer(BASE) + CurOption -1
      Else Begin
        Result := -1;
      End;
      Done := True;
    End Else

      // Is this key the same as the last one?

      If KeyChar = LASTKEY Then Begin

        // Yes - make it repeat if necessary.

        If FRAMES - REPCOUNT >= LongWord(RepeatLen) Then Begin
          RepeatLen := REPPER;
          REPCOUNT := FRAMES;
          Case LASTKEY of
            K_UP:
              If CurOption > 1 Then Dec(CurOption) Else CurOption := Numoptions -1;
            K_DOWN:
              If CurOption < NumOptions -1 Then Inc(CurOption) Else CurOption := 1;
            K_HOME:
              CurOption := 1;
            K_END:
              CurOption := NumOptions -1;
          End;
          SP_PlaySystem(CLICKCHAN, CLICKBANK);
        End;

      End Else Begin

        // No - This is a new key

        Case LASTKEY of
          K_UP:
            If CurOption > 1 Then Dec(CurOption) Else CurOption := Numoptions -1;
          K_DOWN:
            If CurOption < NumOptions -1 Then Inc(CurOption) Else CurOption := 1;
          K_HOME:
            CurOption := 1;
          K_END:
            CurOption := NumOptions -1;
        End;
        SP_PlaySystem(CLICKCHAN, CLICKBANK);
        RepeatLen := REPDEL;
        REPCOUNT := FRAMES;
        KeyChar := LASTKEY;

      End;

  Until Done;

  SP_SetDrawingWindow(SCREENBANK, Window);
  SP_DeleteWindow(WindowID, Err);
  OptionList.Free;

End;

end.
