unit SP_Math;

interface

Uses Math, SP_Util, SP_Errors, SP_Variables;

Function SP_Polyterm(x, p, q, r: aFloat; n: Integer): aFloat;
Function SP_Under(x, p, q, r: aFloat; n: Integer): aFloat;
Function SP_CompSimpson(fn: aString; a, b: aFloat; n: Integer; Var Error: TSP_ErrorCode): aFloat;
Function Mandel(x, y: aFloat; MaxIters: Integer): Integer;

implementation

Uses SP_Interpret_PostFix;

Function SP_Polyterm(x, p, q, r: aFloat; n: Integer): aFloat;
Var
  acc, mul, harm, zeta, corr: aFloat;
  start: Integer;
Begin

  acc := 0;
  mul := 1;

  For start := 1 to n-1 Do Begin

    harm := ln(start);
    If r = 0 Then zeta := 1 Else zeta := exp(harm * r);
    If p = 0 Then harm := 1 Else harm := exp(harm * p);
    mul := mul * x * harm;
    acc := acc + mul * zeta * cos(q * (start - 1));

  End;

  Result := acc;

End;

Function SP_Under(x, p, q, r: aFloat; n: Integer): aFloat;
Var
  acc, harm: aFloat;
  start: Integer;
Begin

  acc := 0;
  For start := 1 To n -1 Do Begin

    harm := x / Start;
    acc := acc + harm + SP_Polyterm(harm, p, q, r, n);

  End;

  Result := SP_Polyterm(acc, r, q, p, n);

End;

Function SP_CompSimpson(fn: aString; a, b: aFloat; n: Integer; Var Error: TSP_ErrorCode): aFloat;
Var
  h, x, S, F, FnVal: aFloat;
  i, Idx, Idx2, VarOffsetN, ValPosition: Integer;
  ValTkn: paString;
  ParamName: aString;

  Function ExecuteFN: aFloat;
  Begin

    // Executes a function (as in, a DEF FN function in SpecBAS code) set up
    // in the parent proc.

    VarOffsetN := NVLen;
    Idx2 := SP_NewNumVar;
    NumVars[Idx2]^.Content.VarType := SP_SIMPLE;
    NumVars[Idx2]^.Name := ParamName;
    NumVars[Idx2]^.ProcVar := True;
    NumVars[Idx2]^.Content.Value := FnVal;
    NumVars[Idx2]^.ContentPtr := @NumVars[Idx2]^.Content;

    ValPosition := 1;
    SP_Interpret(ValTkn, ValPosition, Error);

    Result := SP_StackPtr^.Val;
    Dec(SP_StackPtr);

    SP_ResizeNumVars(VarOffsetN);

  End;

Begin

  // A more complex function - requires a function to be executed. Find that function now:
  // Must be numeric, and contain one numeric parameter.

  Idx := 0;
  While Idx < Length(SP_FnList) Do Begin

    If SP_FnList[Idx].Name = fn Then
      Break
    Else
      Inc(Idx);

  End;

  If Idx < Length(SP_FnList) Then Begin

    // Check the parameter count - must be one.

    If SP_FnList[Idx].ParamCount <> 1 Then Begin

      Error.Code := SP_ERR_PROC_PARAM_COUNT;

    End Else Begin

      // Now check that it's a numeric parameter

      If pFnVar(@SP_FnList[Idx].ParamList[1])^.ID = 1 Then Begin

        Error.Code := SP_ERR_PARAMETER_ERROR;
        Exit;

      End Else Begin

        ParamName := Copy(SP_FnList[Idx].ParamList, 6, Length(SP_FnList[Idx].ParamList));
        ValTkn := @SP_FnList[Idx].Expr;

      End;

    End;

  End Else Begin

    Error.Code := SP_ERR_FN_NOT_FOUND;
    Exit;

  End;

  h := (b - a) / (2 * n);
  FnVal := a;
  S := ExecuteFn;

  If Error.Code <> SP_ERR_OK Then Exit;

  i := 1;
  While I < 2 * n Do Begin
    x := a + h * i;
    FnVal := x;
    s := s + 4 * ExecuteFn;
    If Error.Code <> SP_ERR_OK Then Exit;
    Inc(i, 2);
  End;

  i := 2;
  While I < 2 * n -1 Do Begin
    x := a + h * i;
    FnVal := x;
    S := S + 2 * ExecuteFn;
    If Error.Code <> SP_ERR_OK Then Exit;
    Inc(i, 2);
  End;

  FnVal := b;
  S := S + ExecuteFn;
  Result := h * S / 3;

End;

Function Mandel(x, y: aFloat; MaxIters: Integer): Integer;
Var
  p2y, q, zr, zi, ckr, cki, tmp: aFloat;
  p, ptot: Integer;
Begin

  Result := 0;

  p2y := y * y;
  q := ((x - 0.25) * (x - 0.25)) + p2y;
  If (((x + 1)*(x + 1)) + p2y < 0.0625) or (q * (q + (x - 0.25)) < p2y / 4) Then
    Exit;

  zr := x;
  zi := y;
  p := 0;
  ptot := 8;

  Repeat

    ckr := zr;
    cki := zi;
    Inc(pTot, pTot);
    If pTot > MaxIters Then
      pTot := MaxIters;

    Repeat

      Inc(p);
      tmp := (zr * zr) - (zi * zi) + x;
      zi := (zi * 2 * zr) + y;
      zr := tmp;

      If (zr * zr) + (zi * zi) > 4 Then Begin
        Result := p;
        Exit;
      End Else
        If (zr = ckr) And (zi = cki) Then
          Exit;

    Until p >= ptot;

  Until ptot = MaxIters;

End;

end.


